#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **application dependency** (i.e., both mandatory and optional
third-party Python packages required by this application) facilities.

This low-level submodule defines functions intended to be called by high-level
submodules (e.g., :mod:`betse.util.cli.cliabc`) *before* attempting to import
any such dependencies.
'''

# ....................{ TODO                               }....................
#FIXME: Refactor this entire submodule away from the deprecated
#"setuptools.pkg_resources" API to the standard "importlib.metadata" API. *sigh*
#
# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To raise human-readable exceptions on missing mandatory
# dependencies, the top-level of this module may import *ONLY* from packages
# guaranteed to exist at initial runtime (i.e., standard Python and application
# packages).
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

from betse.exceptions import BetseLibException
# from betse.util.io.log import logs
# from betse.util.app.meta import appmetaone
from betse.util.py.module.pymodname import (
    DEPENDENCY_TO_COMMANDS,
    DEPENDENCY_TO_MODULE_NAME,
    import_module,
    is_module,
)
# from betse.util.type.iterable import itertest
from betse.util.type.types import (
    # MappingType,
    ModuleOrSequenceTypes,
    # SequenceTypes,
)

# ....................{ EXCEPTIONS                         }....................
def die_unless_runtime_mandatory_all() -> None:
    '''
    Raise an exception unless all mandatory runtime dependencies of this
    application are **satisfiable** (i.e., both importable and of a
    satisfactory version) *and* all external commands required by these
    dependencies (e.g., GraphViz's ``dot`` command) reside in the current
    ``${PATH}``.

    Equivalently, this function raises an exception if at least one such
    dependency is unsatisfied. For importable unsatisfied dependencies with
    :mod:`setuptools`-specific metadata (e.g., ``.egg-info/``-suffixed
    subdirectories of the ``site-packages/`` directory for the active Python 3
    interpreter, typically created by :mod:`setuptools` at install time), this
    function additionally validates the versions of these dependencies to
    satisfy all application requirements.

    Raises
    ------
    BetseLibException
        If at least one mandatory runtime dependency is unsatisfiable.
    '''

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Application-wide dependency metadata submodule.
    # metadeps = appmetaone.get_app_meta().module_metadeps
    #
    # # If at least one passed dependency is unsatisfied, raise an exception.
    # die_unless_requirements_dict(metadeps.RUNTIME_MANDATORY)
    pass


def die_unless_runtime_optional(*dependency_names: str) -> None:
    '''
    Raise an exception unless all optional runtime dependencies of this
    application with the passed :mod:`setuptools`-specific project names are
    **satisfiable** (i.e., both importable and of a satisfactory version)
    *and* all external commands required by these dependencies (e.g.,
    GraphViz's ``dot`` command) reside in the current ``${PATH}``.

    Parameters
    ----------
    dependency_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`metadeps.RUNTIME_OPTIONAL` dictionary), an exception is raised.

    Raises
    ------
    BetseLibException
        If at least one such dependency is unsatisfiable.

    See Also
    --------
    :func:`die_unless_runtime_mandatory_all`
        Further details.
    '''

    # If one or more of these optional runtime dependencies are unsatisfied...
    if not is_runtime_optional(*dependency_names):
        # For the name of each such dependency...
        for dependency_name in dependency_names:
            # If this optional runtime dependencies is unsatisfied, raise an
            # exception.
            if not is_runtime_optional(dependency_name):
                #FIXME: Non-ideal. We should explicitly inform the user *WHY*
                #this dependency is unsatisfies. Good enough for now, though!
                raise BetseLibException(
                    f'Dependency "{dependency_name}" unsatisfied.')
            # Else, this optional runtime dependencies is satisfied.
    # Else, *ALL* of these optional runtime dependencies are satisfied.

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Application-wide dependency metadata submodule.
    # metadeps = appmetaone.get_app_meta().module_metadeps
    #
    # # If at least one passed dependency is unsatisfied, raise an exception.
    # die_unless_requirements_dict_keys(
    #     metadeps.RUNTIME_OPTIONAL, *dependency_names)

# ....................{ EXCEPTIONS ~ dict                  }....................
#FIXME: Preserved for temporary posterity. </facepalm>
# @type_check
# def die_unless_requirements_dict(requirements_dict: MappingType) -> None:
#     '''
#     Raise an exception unless all dependencies described by the passed
#     dictionary are **satisfiable** (i.e., both importable and of a satisfactory
#     version) *and* all external commands required by these dependencies (e.g.,
#     GraphViz's ``dot`` command) reside in the current ``${PATH}``.
#
#     Parameters
#     ----------
#     requirements_dict : MappingType
#         Dictionary mapping from the names of all :mod:`setuptools`-specific
#         projects implementing these dependencies to the requirements strings
#         constraining these dependencies.
#
#     Raises
#     ----------
#     BetseLibException
#         If at least passed dependency is unsatisfiable.
#     '''
#
#     # Import the following submodule, which globally imports "pkg_resources",
#     # *AFTER* validating "pkg_resources" to be importable.
#     from betse.lib.setuptools import setuptool
#
#     # Validate these dependencies.
#     setuptool.die_unless_requirements_dict(requirements_dict)
#
#     # Validate all external commands required by these dependencies.
#     die_unless_command(*requirements_dict.keys())
#
#
# @type_check
# def die_unless_requirements_dict_keys(
#     requirements_dict: MappingType, *dependency_names: str) -> None:
#     '''
#     Raise an exception unless all dependencies with the passed
#     :mod:`setuptools`-specific project names described by the passed dictionary
#     are **satisfiable** (i.e., both importable and of a satisfactory version)
#     *and* all external commands required by these dependencies (e.g.,
#     GraphViz's ``dot`` command) reside in the current ``${PATH}``.
#
#     Parameters
#     ----------
#     requirements_dict : MappingType
#         Dictionary mapping from the names of all :mod:`setuptools`-specific
#         projects implementing these dependencies to the requirements strings
#         constraining these dependencies.
#     dependency_names : tuple[str]
#         Tuple of the names of all :mod:`setuptools`-specific projects
#         implementing these dependencies (e.g., ``NetworkX``). If any such
#         name is *not* a key of this dictionary, an exception is raised.
#
#     Raises
#     ----------
#     BetseLibException
#         If at least passed dependency is unsatisfiable.
#     '''
#
#     # Avoid circular import dependencies.
#     from betse.lib.setuptools import setuptool
#
#     # Validate these dependencies.
#     setuptool.die_unless_requirements_dict_keys(
#         requirements_dict, *dependency_names)
#
#     # Validate all external commands required by these dependencies.
#     die_unless_command(*dependency_names)
#
# # ....................{ EXCEPTIONS ~ commands             }....................
# @type_check
# def die_unless_command(*dependency_names: str) -> None:
#     '''
#     Raise an exception unless all external commands required by all application
#     dependencies (of any type, including optional, mandatory, runtime, testing,
#     or otherwise) with the passed :mod:`setuptools`-specific project names are
#     **installed** (i.e., are executable files in the current ``${PATH}``).
#
#     Parameters
#     ----------
#     dependency_names : Tuple[str]
#         Tuple of the names of all :mod:`setuptools`-specific projects
#         corresponding to these dependencies (e.g., ``NetworkX``).
#
#     Raises
#     ----------
#     BetseLibException
#         If any external command required by any such dependency is *not* found.
#     '''
#
#     # Avoid circular import dependencies.
#     from betse.util.os.command import cmdpath
#
#     # If any external command required by any such dependency is *NOT* found,
#     # iteratively search for the first such missing command and raise a
#     # human-readable exception synopsizing this command. For efficiency, this
#     # inefficient iteration is performed *ONLY* as required.
#     if not is_command(*dependency_names):
#         # For the name of each such dependency...
#         for requirement_name in dependency_names:
#             # For each "RequirementCommand" instance describing an external
#             # command required by this dependency...
#             for dependency_command in _iter_requirement_commands(
#                 requirement_name):
#                 # If this command is *NOT* in the ${PATH}, raise an exception.
#                 if not cmdpath.is_pathable(dependency_command.basename):
#                     raise BetseLibException(
#                         'Dependency "{}" unsatisfied, as '
#                         '{} not installed '
#                         '(i.e., command "{}" not found in ${{PATH}}).'.format(
#                             requirement_name,
#                             dependency_command.name,
#                             dependency_command.basename,
#                         ))

# ....................{ TESTERS                           }....................
def is_commands(*dependency_names: str) -> bool:
    '''
    ``True`` only if all external commands required by all application
    dependencies (of any type, including optional, mandatory, runtime, testing,
    or otherwise) with the passed :mod:`setuptools`-specific project names are
    **installed** (i.e., are executable files in the current ``${PATH}``).

    Parameters
    ----------
    dependency_names : Tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        corresponding to these dependencies (e.g., ``NetworkX``).
    '''

    # Avoid circular import dependencies.
    from betse.util.os.command.cmdpath import is_pathable

    # Return true only if...
    return all(
        # Each external command required by each dependency is in the ${PATH}.
        is_pathable(dependency_command.basename)
        # For the name of each passed dependency...
        for dependency_name in dependency_names
        # For each "RequirementCommand" instance describing an external command
        # required by this dependency...
        for dependency_command in DEPENDENCY_TO_COMMANDS.get(
            dependency_name, ())
    )

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Return true only if...
    # return all(
    #     # Each external command required by each dependency is in the ${PATH}.
    #     cmdpath.is_pathable(dependency_command.basename)
    #     # For the name of each passed dependency...
    #     for requirement_name in dependency_names
    #     # For each "RequirementCommand" instance describing an external command
    #     # required by this dependency...
    #     for dependency_command in _iter_requirement_commands(requirement_name)
    # )


def is_runtime_optional(*dependency_names: str) -> bool:
    '''
    ``True`` only if all optional runtime dependencies of this application with
    the passed :mod:`setuptools`-specific project names are **satisfiable**
    (i.e., both importable and of a satisfactory version) *and* all external
    commands required by these dependencies (e.g., GraphViz's ``dot`` command)
    reside in the current ``${PATH}``.

    Parameters
    ----------
    dependency_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such
        name is *not* a key of the :data:`metadeps.RUNTIME_OPTIONAL` dictionary
        and is thus unrecognized, an exception is raised.

    Returns
    -------
    bool
        ``True`` only if these optional runtime dependencies are satisfiable.

    See Also
    --------
    :func:`die_unless_runtime_mandatory_all`
        Further details.
    '''

    # For the name of each passed optional runtime dependency...
    for dependency_name in dependency_names:
        # Fully-qualified name of the top-level package or module defining this
        # dependency, defaulting to the name of this dependency.
        module_name = DEPENDENCY_TO_MODULE_NAME.get(
            dependency_name, dependency_name)

        # If it is *NOT* the case that...
        if not (
            # This dependency is importable *AND*...
            is_module(module_name) and
            # All external commands required by this dependency are installed in
            # the current "${PATH}"...
            is_commands(dependency_name)
        # Then immediately return false.
        ):
            return False
        # Else, this dependency is importable *AND* all external commands
        # required by this dependency are installed in the current "${PATH}".
        # Silently continue to the next dependency.

    # Return true.
    return True

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Application-wide dependency metadata submodule.
    # metadeps = appmetaone.get_app_meta().module_metadeps
    #
    # # Return true only if these optional runtime dependencies are satisfiable.
    # return is_requirements_dict_keys(
    #     metadeps.RUNTIME_OPTIONAL, *dependency_names)


#FIXME: Preserved for temporary posterity. </facepalm>
# @type_check
# def is_requirements_dict_keys(
#     requirements_dict: MappingType, *dependency_names: str) -> bool:
#     '''
#     ``True`` only if all dependencies with the passed
#     :mod:`setuptools`-specific project names are **satisfiable** (i.e., both
#     importable and of a satisfactory version) *and* all external commands
#     required by these dependencies (e.g., GraphViz's ``dot`` command) reside in
#     the current ``${PATH}``.
#
#     Parameters
#     ----------
#     requirements_dict : MappingType
#         Dictionary mapping from the names of all :mod:`setuptools`-specific
#         projects implementing these dependencies to the requirements strings
#         constraining these dependencies.
#     dependency_names : tuple[str]
#         Tuple of the names of all :mod:`setuptools`-specific projects
#         implementing these dependencies (e.g., ``NetworkX``). If any such
#         name is *not* a key of this dictionary, an exception is raised.
#
#     Returns
#     ----------
#     bool
#         ``True`` only if these dependencies are all satisfiable.
#
#     See Also
#     ----------
#     :func:`die_unless_runtime_mandatory_all`
#         Further details.
#     '''
#
#     # Avoid circular import dependencies.
#     from betse.lib.setuptools import setuptool
#
#     # Tuple of requirements strings converted from this subset of key-value
#     # pairs of these requirements.
#     requirements_tuple = setuptool.get_requirements_str_from_dict_keys(
#         requirements_dict, *dependency_names)
#
#     # Return True only if...
#     return (
#         # These dependencies are all satisfied.
#         setuptool.is_requirement_str(*requirements_tuple) and
#         # All external commands required by these dependencies are installed.
#         is_command(*dependency_names)
#     )

# ....................{ GETTERS ~ metadata                }....................
def get_metadatas() -> tuple:
    '''
    Tuple of 2-tuples `(metedata_name, metadata_value`), describing all
    currently installed optional and mandatory third-party dependencies
    required at both runtime and testing time.
    '''

    # Defer heavyweight imports.
    from betse.lib.matplotlib.matplotlibs import mpl_config
    from betse.lib.numpy import numpys
    # from betse.lib.setuptools import setuptool

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Application-wide dependency metadata submodule.
    # metadeps = appmetaone.get_app_meta().module_metadeps
    #
    # # Tuple of all dependency versions.
    # LIB_VERSION_METADATA = (
    #     # Dependencies metadata.
    #     ('runtime dependencies (mandatory)',
    #      setuptool.get_requirements_dict_synopsis(metadeps.RUNTIME_MANDATORY)),
    #     ('runtime dependencies (optional)',
    #      setuptool.get_requirements_dict_synopsis(metadeps.RUNTIME_OPTIONAL)),
    #     ('testing dependencies (mandatory)',
    #      setuptool.get_requirements_dict_synopsis(metadeps.TESTING_MANDATORY)),
    # )

    # Tuple of all dependency-specific metadata.
    LIB_METADATA = (
        # matplotlib metadata.
        ('matplotlib', mpl_config.get_metadata()),
    )

    # Return a tuple aggregating these tuples.
    return (
        # LIB_VERSION_METADATA +
        LIB_METADATA +
        numpys.get_metadatas()
    )

# ....................{ IMPORTERS                          }....................
def import_runtime_optional(*dependency_names: str) -> ModuleOrSequenceTypes:
    '''
    Import and return either a single top-level module object if passed one
    dependency name *or* sequence of top-level module objects otherwise (i.e.,
    if passed multiple such names).

    Each returned module object is guaranteed to satisfy the optional runtime
    dependency of this application with the corresponding requirement name.

    Parameters
    ----------
    dependency_names : tuple[str]
        Tuple of the names of all :mod:`setuptools`-specific projects
        implementing these dependencies (e.g., ``NetworkX``). If any such name
        is unrecognized (i.e., is *not* a key of the
        :data:`metadeps.RUNTIME_OPTIONAL` dictionary), an exception is raised.
    '''

    # If passed only a single optional runtime dependency...
    if len(dependency_names) == 1:
        # Name of this dependency.
        dependency_name = dependency_names[0]

        # Fully-qualified name of the top-level package or module defining this
        # dependency, defaulting to the name of this dependency.
        module_name = DEPENDENCY_TO_MODULE_NAME.get(
            dependency_name, dependency_name)

        # Dynamically import and return this dependency.
        return import_module(module_name)
    # Else, two or more dependencies were passed.

    # List of all dynamically imported packages and modules to be returned.
    dependencies = []

    # For the name of each such dependency...
    for dependency_name in dependency_names:
        # Fully-qualified name of the top-level package or module defining this
        # dependency, defaulting to the name of this dependency.
        module_name = DEPENDENCY_TO_MODULE_NAME.get(
            dependency_name, dependency_name)

        # Dynamically import and return this dependency.
        dependency = import_module(module_name)

        # Append this dependency to this list.
        dependencies.append(dependency)

    # Return these dynamically imported packages and modules.
    return dependencies

    #FIXME: Resurrect this when refactoring to "importlib.metadata", please.
    # # Application-wide dependency metadata submodule.
    # metadeps = appmetaone.get_app_meta().module_metadeps
    #
    # # Import and return these modules.
    # return import_requirements_dict_keys(
    #     metadeps.RUNTIME_OPTIONAL, *dependency_names)


#FIXME: Preserved for temporary posterity. </facepalm>
# @type_check
# def import_requirements_dict_keys(
#     requirements_dict: MappingType, *dependency_names: str) -> (
#     ModuleOrSequenceTypes):
#     '''
#     Import and return either a single top-level module object if passed one
#     :mod:`setuptools`-specific requirement name *or* sequence of top-level
#     module objects otherwise (i.e., if passed multiple such names).
#
#     Each returned module object is guaranteed to satisfy the dependency of this
#     application with the corresponding requirement name as a key of the passed
#     dictionary.
#     '''
#
#     # Avoid circular import dependencies.
#     from betse.lib.setuptools import setuptool
#
#     # Validate these dependencies.
#     die_unless_requirements_dict_keys(requirements_dict, *dependency_names)
#
#     # Validate all external commands required by these dependencies.
#     return setuptool.import_requirements_dict_keys(
#         requirements_dict, *dependency_names)



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests for the :mod:`betse.util.path.archives` submodule.
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse.util.test.pytest.mark.pytskip import skip_unless_module

# ....................{ CONSTANTS                          }....................
# Ideally, the existing
# "betse.util.path.archives._ARCHIVE_FILETYPE_TO_MODULE_NAME" dictionary would
# be leveraged to populate this parametrization. Unfortunately, doing so would
# necessitate unsafely importing the "betse.util.path.archives" submodule
# above, ensuring failure of the entire test suite on the inability to import
# this submodule -- absurd and frankly unacceptable test-time fragility.
# Instead, the set of all archive filetypes is redefined in test-oriented
# manner ignoring filetypes whose corresponding optional stdlib modules are
# unavailable.
#
# See the "_ARCHIVE_FILETYPE_TO_MODULE_NAME" dictionary for further details.
ARCHIVE_FILETYPES = (
    # The higher-level skip_unless_lib_runtime_optional() decorator is
    # intentionally *NOT* called here. Although the following modules could be
    # argued to be optional runtime dependencies of this application requiring
    # addition to the "betse.metadata.RUNTIME_OPTIONAL" dictionary,
    # these modules are pre-packaged with Python itself rather than installed
    # via setuptools. Since these modules declare no "__version__" attribute,
    # validating the satisfiability of these modules is infeasible -- unlike
    # with customary setuptools-installed optional dependencies.
    #
    # The lower-level skip_unless_module() decorator thus applies here.
    pytest.param('bz2', marks=skip_unless_module('bz2')),
    pytest.param('gz',  marks=skip_unless_module('gzip')),
    pytest.param('xz',  marks=skip_unless_module('lzma')),
)
'''
Tuple of all supported archive filetypes, ignoring those whose corresponding
optional stdlib modules are unimportable under the active Python interpreter
and hence *not* installed at Python installation time.
'''


ARCHIVE_BYTES = bytes(
    (
        "The economics of the future is somewhat different. "
        "You see, money doesn't exist in the 24th century. "
        "The acquisition of wealth is no longer "
        "the driving force in our lives. "
        "We work to better ourselves and the rest of Humanity."
    ), encoding='utf-8',
)
'''
Arbitrary sequence of bytes to be archived.
'''

# ....................{ TESTS                              }....................
@pytest.mark.parametrize(('filetype',), ARCHIVE_FILETYPES,)
def test_archives_read_write_bytes(
    betse_temp_dir: 'py._path.local.LocalPath', filetype: str) -> None:
    '''
    Unit test both the :func:`reading_bytes` and :func:`writing_bytes`
    functions of the :mod:`betse.util.path.archives` submodule for the passed
    archive filetype guaranteed to be supported by the active Python
    interpreter.

    Parameters
    ----------
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to the current
        test.
    filetype : str
        Filetype of the archive format to be tested (e.g., "bz2", "gz", "zip").
    '''

    # Defer heavyweight imports.
    from betse.util.path import archives

    # Absolute path of an archive file of arbitrary basename and the passed
    # filetype in this temporary directory.
    archive_filepath = betse_temp_dir.join('memory_alpha.' + filetype)

    # This path as a string rather than "LocalPath" object.
    archive_filename = str(archive_filepath)

    # Ensure this archive filetype to be supported.
    assert archives.is_filetype(archive_filename)

    # Create this archive.
    with archives.write_bytes(archive_filename) as archive_writer:
        archive_writer.write(ARCHIVE_BYTES)

    # Assert this archive to have been created.
    assert archive_filepath.check(file=1)

    # Unpack this archive.
    with archives.read_bytes(archive_filename) as archive_reader:
        archive_bytes_read = archive_reader.read()

    # Assert this archive to contain the bytes written to that archive.
    assert ARCHIVE_BYTES == archive_bytes_read



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level logging configuration functionality.
'''

# ....................{ IMPORTS                            }....................
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
# WARNING: To avoid circular import dependencies, avoid importing from *ANY*
# application-specific modules at the top-level -- excluding those explicitly
# known *NOT* to import from this module. Since all application-specific
# modules should *ALWAYS* be able to safely import from this module at any
# scoping level, circularities are best avoided here rather than elsewhere.
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
from betse.exceptions import BetseLogException
from betse.util.type.types import type_check

# ....................{ GLOBALS                            }....................
_log_conf = None
'''
Singleton logging configuration for the current Python process.

This configuration provides access to root logger handlers. In particular, this
simplifies modification of logging levels at runtime (e.g., in response to
command-line arguments or configuration file settings).
'''

# ....................{ EXCEPTIONS                         }....................
def die_unless_log_conf() -> None:
    '''
    Raise an exception unless a logging configuration already exists.

    Equivalently, this function raises an exception if *no* logging
    configuration currently exists.

    Raises
    ----------
    BetseLogException
        If no singleton logging configuration currently exists, typically due
        to this function having been called either:

        * *Before* the startup-time call to the :func:`init` function.
        * *After* the shutdown-time call to the :func:`deinit` function.
    '''

    # If no logging configuration currently exists, raise an exception.
    if _log_conf is None:
        raise BetseLogException(
            'Logging unconfigured (e.g., due to either '
            'logconf.init() not having been called or '
            'logconf.deinit() having been called).'
        )

# ....................{ INITIALIZERS                       }....................
@type_check
def init() -> None:
    '''
    Enable our default logging configuration for the active Python process if
    this configuration has yet to be enabled *or* reduce to a noop otherwise
    (e.g., if this method has already been called).

    Specifically, this function instantiates the private :data:`_log_conf`
    singleton to an instance of the application-specific :class:`LogConf`
    class, publicly accessible via the :func:`get_log_conf` module getter. This
    singleton defines sane default filters, formatters, and handlers for the
    root logger, which callers may customize by setting singleton properties.
    '''

    # Avoid circular import dependencies.
    from betse.util.io.log import logs
    from betse.util.io.log.conf.logconfcls import LogConf

    # Module-scoped variables to be set below.
    global _log_conf

    # If a logging configuration already exists...
    if _log_conf is not None:
        # Log a non-fatal warning.
        logs.log_warning(
            'Logging already configured (e.g., due to '
            'logconf.init() having been called).'
        )

        # Reduce to a noop.
        return

    # Instantiate this singleton global with the requisite defaults.
    # print('Reinitializing logging.')
    _log_conf = LogConf()

    # Log this initialization *AFTER* guaranteeing logging sanity.
    logs.log_debug('Initialized singleton logging configuration.')


@type_check
def deinit() -> None:
    '''
    Disable our default logging configuration for the active Python process if
    this configuration has yet to be disabled *or* reduce to a noop otherwise
    (e.g., if this method has already been called).

    Specifically, this function calls the :meth:`LogConf.deinit` method of the
    private :data:`_log_conf` singleton and then nullifies that singleton.
    '''

    # Avoid circular import dependencies.
    from betse.util.io.log import logs

    # Module-scoped variables to be set below.
    global _log_conf

    # If no logging configuration exists, silently noop.
    #
    # Note that this common edge case occurs when an exception is raised early
    # at application startup *BEFORE* the associated init() function is called.
    # Ergo, this constitutes neither a fatal error nor non-fatal warning.
    if _log_conf is None:
        return
    # Else, a logging configuration currently exists.

    # Log this deinitialization *AFTER* guaranteeing logging sanity.
    logs.log_debug('Deinitializing singleton logging configuration...')

    # Deinitialize this logging configuration.
    _log_conf.deinit()

    # Nullify this singleton global for safety *AFTER* all other actions above.
    _log_conf = None

# ....................{ GETTERS                            }....................
def get_log_conf() -> 'betse.util.io.log.conf.logconfcls.LogConf':
    '''
    Singleton logging configuration for the active Python process.

    Raises
    ------
    BetseLogException
        If no singleton logging configuration currently exists.
    '''

    # If no logging configuration exists, raise an exception.
    die_unless_log_conf()

    # Else, a logging configuration exists. Return this configuration.
    return _log_conf


def get_metadata():
    '''
    Ordered dictionary synopsizing the current logging configuration.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

    # Return this dictionary.
    return OrderedArgsDict(
        'file', _log_conf.filename,
        'file level', _log_conf.file_level.name.lower(),
        'verbose', str(_log_conf.is_verbose).lower(),
    )



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level floating point facilities.
'''

# ....................{ IMPORTS                            }....................
import math
from sys import float_info
# from betse.util.io.log import logs
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.types import type_check, RegexCompiledType
from decimal import Decimal

# ....................{ CONSTANTS ~ pi                      }....................
PI = math.pi
'''
**π** (i.e., transcendental number defined as the ratio of a circle's
circumference C to diameter d: ``C/d``).
'''


TWO_PI = 2 * PI
'''
**2π** (i.e., 2 * :data:`PI`), preserved as a global constant to reduce
needless recomputation of this frequently referenced constant.
'''

# ....................{ CONSTANTS ~ range                  }....................
FLOAT_MIN = float_info.min
'''
Minimum representable finite floating point number under the active Python
interpreter.
'''


FLOAT_MAX = float_info.max
'''
Maximum representable finite floating point number under the active Python
interpreter.
'''

# ....................{ TESTERS                            }....................
@type_check
def is_float_str(text: str) -> bool:
    '''
    ``True`` only if the passed string syntactically conforms to either the
    decimal format (e.g., ``6.669``) or scientific notation format (e.g.,
    ``6.66e9``) expected by floating point numbers.

    Equivalently, this tester returns ``True`` only if this string is losslessly
    convertable into a floating point number.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Return True only if this string matches a floating point format.
    return regexes.is_match(text=text, regex=get_float_regex())

# ....................{ GETTERS ~ base 10                  }....................
@type_check
def get_base_10_exponent(number: float) -> int:
    '''
    **Base-10 exponent** (i.e., integer power of ten to which raising a
    corresponding real mantissa yields a floating point number) of the passed
    floating point number.

    Equivalently, this function returns the quantity ``exponent`` in the
    following equation, where ``number`` is the passed number, ``mantissa`` is
    this number's mantissa, and ``exponent`` is this number's exponent:

    .. code:: python

       number = mantissa * 10**exponent

    Examples
    ----------
        >>> from betse.util.type.numeric import floats
        >>> floats.get_base_10_exponent(110001000000000009)
        17
        >>> floats.get_base_10_exponent(1)
        1
        >>> floats.get_base_10_exponent(0.1)
        -1
        >>> floats.get_base_10_exponent(0.110001000000000009)
        -1

    See Also
    ----------
    https://stackoverflow.com/a/45359185/2809027
        StackOverflow post strongly inspiring this implementation.
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # NOTE: This function is intentionally implemented in terms of the
    # high-level "Decimals" class rather than low-level primitives (e.g.,
    # "float", "str"). Doing so necessitates converting from the latter to the
    # former, which necessarily introduces IEEE 754 rounding errors. While these
    # errors are inexcusable in most cases, they are entirely ignorable in this
    # case. Why? Because the exponent of a floating point number is a
    # sufficiently coarse-grained quantity that no accumulation of these errors
    # is likely to permute this quantity. That is, the exponent is robust
    # against rounding errors. While alternatives to a "Decimals"-based approach
    # do technically exist, their complexity and hence fragility warrants the
    # terseness of the current approach.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Base-10 fixed-point number equivalent to this floating point number.
    number_fixed = Decimal(number)

    # Components of this fixed-point number:
    #
    # * "sign", either:
    #   * 0 if this number is positive.
    #   * 1 if this number is negative.
    # * "digits", a tuple of all base-10 digits comprising this number.
    # * "exponent", the integer power of ten to which raising the integer
    #   produced by concatenating the digits of "digits" yields this number.
    #
    # This number is thus given by:
    #
    #     number = (-1)**sign * digits * 10**exponent
    #
    # For the purposes of calculating the desired exponent to be returned, note
    # that the the "sign" term is safely ignorable. Ergo, "f" reduces to:
    #
    #     number = digits * 10**exponent
    #
    # Further note that the form of this expression is equivalent to that of the
    # similar expression in this function's docstring. The difference is that:
    #
    # * The "digits" term is a non-negative integer (i.e., >= 0).
    # * The "mantissa" term in that similar expression is a real number
    #   satisfying the inequality: "10 < mantissa <= 1".
    #
    # Converting the integer "exponent" term in this expression to the real
    # number "exponent" term in that similar expression thus reduces to the
    # following iterative salgorithm:
    #
    # * If the decimal place in the "digits" term is *NOT* currently to the
    #   right of the leftmost digit in this term:
    #   1. Shift this decimal place left by one digit, thus decreasing this
    #      number by a power of ten.
    #   2. Increment this "exponent" term by one, thus increasing this number by
    #      a power of ten. Since doing so is the inverse of the prior operation,
    #      this number's actual value remains unchanged.
    #
    # Note that the decimal place in this "digits" term begins to the right of
    # the rightmost digit in this term. The number of algorithm iterations (and
    # hence incrementations of this "exponent" term) to be performed is thus
    # the number of digits in this "digits" term minus one.
    #
    # Q.E.D.
    _, digits, exponent_decimals = number_fixed.as_tuple()

    # Base-10 exponent of this floating point number, converted from the
    # base-10 exponent of this fixed-point number via the trivial incrementation
    # scheme given above. And now you know why we use the "Decimals" class.
    exponent_float = exponent_decimals + len(digits) - 1

    # Return this exponent.
    return exponent_float


@type_check
def get_base_10_precision(number: float) -> int:
    '''
    **Base-10 precision** (i.e., length of a floating point number's significand
    excluding leading hidden digit 1) of the passed floating point number.

    Equivalently, this function returns the number of base-10 digits in the
    fractional segment of this number *after* accounting for approximation
    errors in floating point arithmetic.

    Examples
    ----------
        >>> from betse.util.type.numeric import floats
        >>> floats.get_base_10_precision(0)
        0
        >>> floats.get_base_10_precision(0.1)
        1
        >>> floats.get_base_10_precision(0.110001000000000009)
        17
    '''

    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # WARNING: This function is intentionally implemented in terms of low-level
    # primitives (e.g., "float", "str") rather than higher-level classes --
    # notably, "Decimals". While this function could superficially be
    # reimplemented in terms of the latter rather than the former, doing so
    # would return blatantly incorrect and overly large precisions for arbitrary
    # floating point numbers due to IEEE 754 rounding errors. Ergo, the current
    # approach actually is the sane approach (as insane as it inarguably is).
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # String formatted from this number, guaranteed by Python internals to
    # comply with one of the following two formats:
    #
    # * Decimal notation (e.g., "3.1415").
    # * Scientific notation (e.g., "3.1415e+92", "3.141592e-65").
    number_text = str(number)

    # Scientific notation match groups if this number is in scientific notation,
    # capturing the significand and exponent of this number into these groups.
    number_text_groups = regexes.get_match_groups_named(
        text=number_text, regex=get_float_regex())

    # If this number is in decimal notation...
    if number_text_groups['exponent'] is None:
        # Fractional part of this number if any *OR* the empty string otherwise.
        # Retrieving this substring is complicated by the fact that one of two
        # alternate match groups may yield this substring at a time.
        significand_frac = (
            number_text_groups['significand_frac_empty'] or
            number_text_groups['significand_frac_nonempty'] or '')
        # logs.log_debug('significand (fractional): %s', significand_frac)
        # print('significand (fractional): %s' % significand_frac)

        # This number's precision is the number of digits in this part.
        precision = len(significand_frac)
    # Else, this number is in scientific notation. In this case..
    else:
        # print('exponent: %s' % number_text_groups['exponent'])
        # This number's precision is the absolute value (i.e., ignoring the
        # sign) of this number's exponent in this notation.
        precision = abs(int(number_text_groups['exponent']))

    # Return this precision.
    return precision

# ....................{ GETTERS ~ regex                    }....................
# For efficiency in downstream clients (e.g., BETSEE) frequently calling the
# is_ml() function and hence requiring this expression be compiled, this
# expression is intentionally pre-compiled rather than uncompiled and thus
# returned as a cached getter.

@func_cached
def get_float_regex() -> RegexCompiledType:
    '''
    Compiled regular expression matching a floating point number represented as
    a string in either decimal notation (e.g., ``6.69``) *or* scientific
    notation (e.g., ``6.6e9``).

    This expression captures the following named match groups (in order):

    1. ``significand``, yielding the mandatory significand of this number
       *including* optional sign prefix (e.g., ``-6.7`` given ``-6.7e-9``).
    1. ``significand_nonfrac_empty``, yielding the optional non-fractional
       digits of this significand *including* optional sign prefix (e.g., ``-6``
       given ``-6.7e-9``). This group's value is ``None`` for numbers prefixed
       by ``.`` rather than a digit (e.g., ``-.67e-9``).
    1. ``significand_frac_nonempty``, alternatively yielding the optional
       fractional digits of this significand in a manner guaranteed to be
       non-empty (e.g., ``None`` given ``-6.7e-9``).
    1. ``significand_frac_empty``, alternatively yielding the optional
       fractional digits of this significand in a manner *not* guaranteed to be
       non-empty (e.g., ``7`` given ``-6.7e-9``). This edge case is required to
       support floating point numbers of both the form
       ``{significand_nonfrac}.{significand_frac_empty}`` *and*
       ``.{significand_frac_nonempty}``. Since the standard :mod:`re` module
       does *not* support the ``(?|...)``-style branch reset syntax supported by
       the third-party :mod:`regex` module, this is the best we can do without
       burdening the codebase with *yet another* third-party dependency.
    1. ``exponent``, yielding the optional exponent of this number *including*
       optional sign prefix (e.g., ``-9`` given ``6.7e-9``).

    See Also
    ----------
    https://jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html
        Blog article partially inspiring this implementation.
    https://stackoverflow.com/a/658662/2809027
        StackOverflow answer partially inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Create, return, and cache this expression.
    return regexes.compile_regex(
        # Significand (captured).
        r'(?P<significand>'
            # Negative or positive prefix (optional).
            r'[+-]?'
            # Either:
            r'(?:'
            # * Significand prefixed by one or more non-fractional digits.
                # Non-fractional digits.
                r'(?P<significand_nonfrac_empty>\d+)'
                # Fractional digits (optional).
                r'(?:\.(?P<significand_frac_empty>\d*))?'
            r'|'
            # * Significand containing only fractional digits.
                r'\.(?P<significand_frac_nonempty>\d+)'
            r')'
        r')'
        # Exponent (optional).
        r'(?:'
            # Exponent prefix.
            r'[eE]'
            # Exponent (captured).
            r'(?P<exponent>'
                # Exponent sign (optional).
                r'[+-]?'
                # Exponent digits.
                r'\d+'
            r')'
        r')?'
    )


@func_cached
def _get_float_exponent_regex() -> RegexCompiledType:
    '''
    Compiled regular expression matching the exponent of a floating point number
    represented as a string in scientific notation, typically produced by the
    ``{:g}`` format specifier (e.g., ``6.6e+09``).

    This expression captures the following named match groups (in order):

    1. ``negation``, yielding the optional negative sign prefixing this number's
       exponent if any or the empty string otherwise (e.g., ``-`` given
       ``6.7e-09``).
    1. ``magnitude``, yielding the mandatory magnitude of this number's exponent
       excluding optional prefixing zero (e.g., ``9`` given ``6.7e+09``).

    See Also
    ----------
    https://jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html
        Blog article partially inspiring this implementation.
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Create, return, and cache this expression.
    return regexes.compile_regex(
        # Exponent prefix.
        r'e'
        # Exponent sign, either:
        r'(?:'
            # Ignorable positive sign. Since exponents lacking an explicit
            # sign default to positive, this sign is extraneous for purposes
            # of producing human-readable strings from floats.
            r'\+'
            r'|'
            # Non-ignorable negative sign.
            r'(?P<negation>-)'
        r')'
        # Exponent digits.
        #
        # Ignorable zero prefix (optional). For unclear reasons, the "{:g}"
        # format specifier prefixes all exponnet magnitudes in the
        # single-digit range [1, 9] with a zero, which is extraneous for
        # purposes of producing human-readable strings from floats.
        r'0?'
        # Non-ignorable exponent magnitude.
        r'(?P<magnitude>\d+)'
        # Exponent end.
        r'$'
    )

# ....................{ CONVERTERS                         }....................
@type_check
def to_str(number: float) -> str:
    '''
    Human-readable string losslessly converted from the passed floating point
    number.

    This function effectively pretty-prints floats, reducing the unnecessarily
    verbose strings produced by the ``{:g}`` format specifier as follows:

    * The extraneous positive sign preceding positive exponents is removed
      (e.g., reducing ``6.66e+77`` to simply ``6.66e77``).
    * The extraneous zero preceding single-digit exponents is removed (e.g.,
      reducing ``6.66e-07`` to simply ``6.66e-7``).

    Parameters
    ----------
    number : float
        Floating point number to be stringified.

    Returns
    ----------
    str
        Human-readable string losslessly converted from this number.

    See Also
    ----------
    https://jdreaver.com/posts/2014-07-28-scientific-notation-spin-box-pyside.html
        Blog article partially inspiring this implementation.

    Examples
    ----------
        >>> from betse.util.type.numeric import floats
        >>> print(floats.to_str(6.66e7))
        6.66e7
        >>> print('{:g}'.format(6.66e7))
        6.66e+07
    '''

    # Avoid circular import dependencies.
    from betse.util.type.text import regexes

    # Mostly human-readable string losslessly converted from this number.
    number_str = '{:g}'.format(number)

    # Improve this string's readability by removing all extraneous syntax.
    number_str = regexes.replace_substrs(
        text=number_str,
        regex=_get_float_exponent_regex(),

        # Callable passed the object matching the exponent of this floating
        # point number if any. Ideally, a raw string resembling the following
        # would be passed:
        #
        #     replacement=r'e\g<negation>\g<magnitude>',
        #
        # In Python 3.6, this would be feasible. In prior versions of Python,
        # however, this is infeasible. Why? Whereas Python 3.6 expands unmatched
        # backreferences to the empty string, Python < 3.6 expands unmatched
        # backreferences to None, raising the following exception:
        #
        #     sre_constants.error: unmatched group
        replacement=lambda match:
            'e' +
            (match.group('negation') or '') +
            match.group('magnitude')
    )

    # Return this string.
    return number_str



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **warning** (i.e., non-fatal errors with associated types emitted by
the standard :mod:`warnings` module) facilities.
'''

# ....................{ IMPORTS                            }....................
import sys, warnings
from beartype.typing import (
    ContextManager,
    Iterator,
)
from betse.util.io.log import logs
from betse.util.type.types import ClassType
from contextlib import contextmanager

# ....................{ INITIALIZERS                       }....................
def init() -> None:
    '''
    Initialize this submodule.

    Specifically, this function conditionally establishes a default warning
    filter as follows:

    * If the end user explicitly passed the ``-W`` option to the external
      command forking the active Python interpreter (e.g., via ``python3 -W -m
      betse``) and hence expressed one or more warning preferences, defer to
      these preferences as is.
    * Else if this application has a Git-based working tree and is thus likely
      to be under active development, unconditionally log *all* warnings
      (including those ignored by default).
    * Else, preserve the default warning filter defined by the standard
      :mod:`warnings` module.
    '''

    # Avoid circular import dependencies.
    from betse.util.app.meta import appmetaone
    from betse.util.test import tsttest

    # Log this initialization.
    logs.log_debug('Initializing warning policy...')

    # If the end user explicitly passed the "-W" option to the external command
    # forking the active Python interpreter (and hence expressed a warning
    # preference), defer to these preferences.
    if sys.warnoptions:
        logs.log_debug(
            'Deferring warning policy to '
            '"-W" option passed to Python interpreter.')
    # Else if the active Python interpreter is currently exercising tests,
    # defer to the test harness supervising these tests. Most harnesses
    # (including pytest) define sane default warnings filters as well as
    # enabling users to externally configure warnings filters from project-wide
    # configuration files. Ergo, the current harness knows better than we do.
    elif tsttest.is_testing():
        logs.log_debug(
            'Deferring warning policy to that of the parent test harness.')
    # Else...
    else:
        # If this application has a Git-based working tree and is thus likely
        # to be under active development...
        if appmetaone.get_app_meta().is_git_worktree:
            # Log this preference.
            logs.log_debug(
                'Setting warning policy to '
                'unconditionally log all warnings '
                '(i.e., due to detecting developer environment).')

            # Discard all previously registered warnings filter *BEFORE*
            # registering a warnings filter.
            warnings.resetwarnings()

            # Registering a warnings filter unconditionally logging *ALL*
            # warnings, including those Python ignores by default.
            warnings.simplefilter('default')
        # Else, preserve Python's default warning filter as is.
        else:
            logs.log_debug('Deferring to default warning policy.')

# ....................{ MANAGERS                           }....................
def ignoring_deprecations() -> ContextManager:
    '''
    Single-shot context manager temporarily ignoring all **deprecation
    warnings** (i.e., instances of the :class:`DeprecationWarning`,
    :class:`PendingDeprecationWarning`, and :class:`FutureWarning` exception
    base classes) emitted by the :mod:`warnings` module for the duration of
    this context.

    See Also
    -----------
    :class:`ignoring_warnings`
        Further details.
    '''

    return ignoring_warnings(
        DeprecationWarning,
        PendingDeprecationWarning,
        FutureWarning,
    )


@contextmanager
def ignoring_warnings(*warning_clses: ClassType) -> Iterator[None]:
    '''
    Single-shot context manager temporarily ignoring *all* warnings of *all*
    passed warning types emitted by the :mod:`warnings` module for the duration
    of this context.

    Caveats
    -----------
    This context manager is single-shot and hence *not* safely reusable across
    multiple ``with`` blocks. Why? Because the low-level
    :class:`warnings.catch_warnings` class to which this context manager defers
    is itself single-shot, explicitly raising exceptions on reuse. While
    lamentable, this overhead appears to be unavoidable.

    Parameters
    -----------
    warning_clses : Tuple[ClassType]
        Tuple of zero or more warning classes to be ignored by this context
        manager. Defaults to a tuple containing only the :class:`Warning`
        superclass of all warning subclasses (both builtin and user-defined),
        in which case this context manager ignores *all* warnings.
    '''

    # If no types were passed, default to the "Warning" superclass.
    if not warning_clses:
        warning_clses = (Warning,)

    # Isolate all side effects produced by the "warnings" module to this block.
    with warnings.catch_warnings():
        # For each class of warning to ignore...
        for warning_cls in warning_clses:
            # Temporarily ignore all instances of this warning class.
            warnings.simplefilter(action='ignore', category=warning_cls)

        # Yield control to the body of the caller's "with" block.
        yield



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Create and electrodiffuses a suite of customizable general molecule in the
BETSE ecosystem, including functionality to pump the molecule, use it as a
gating ligand, produce and consume it, and use it an enzyme to facilitate
another reaction.

The molecule is assumed to be at low concentrations and to not have a
significant effect on system voltages or currents. This module creates a
structure containing all user-defined molecules, along with the facilities to
initialize, define the core computations for a simulation loop, save and report
on data, and plot.
'''

# --------------------( IMPORTS                           )--------------------
from betse.exceptions import BetsePyDotException
from betse.lib import libs

# --------------------( PLOTTERS                          )--------------------
#FIXME: Ideally, the "self" parameter should be specific to class methods.
#Since these are simple functions with no parent objects, we should probably
#rename all "self" parameters in this submodule to... well, something. Since
#"self" appears to refer to an instance of the "MasterOfGenes" or
#"MasterOfNetworks" classes, "master" might be a suitable parameter name. Yum!
def set_net_opts(self, net_plot_opts, p):

    if net_plot_opts is not None:
        self.plot_net = net_plot_opts['plot network']
        self.net_font_name = net_plot_opts['font name']
        self.node_font_size = net_plot_opts['node font size']
        self.tit_font_size = net_plot_opts['sub-net font size']
        self.net_layout = net_plot_opts['layout']
        self.edge_width = net_plot_opts['edge linewidth']

        self.conc_node_font_color = net_plot_opts['substances']['font color']
        self.conc_node_color = net_plot_opts['substances']['node color']
        self.conc_shape = net_plot_opts['substances']['node shape']

        self.react_node_font_color = net_plot_opts['reactions']['font color']
        self.react_node_color = net_plot_opts['reactions']['node color']
        self.reaction_shape = net_plot_opts['reactions']['node shape']

        self.transp_node_font_color = net_plot_opts['transporters']['font color']
        self.transp_node_color = net_plot_opts['transporters']['node color']
        self.transporter_shape = net_plot_opts['transporters']['node shape']

        self.chan_node_font_color = net_plot_opts['channels']['font color']
        self.chan_node_color = net_plot_opts['channels']['node color']
        self.channel_shape = net_plot_opts['channels']['node shape']

        self.ed_node_font_color = net_plot_opts['electrodiffusion']['font color']
        self.ed_node_color = net_plot_opts['electrodiffusion']['node color']
        self.ed_shape = net_plot_opts['electrodiffusion']['node shape']

        self.vmem_node_font_color = net_plot_opts['Vmem']['font color']
        self.vmem_node_color = net_plot_opts['Vmem']['node color']
        self.vmem_shape = net_plot_opts['Vmem']['node shape']

        self.subnets_dicts = net_plot_opts['sub-nets']
        self.additional_edges = net_plot_opts.get('additional relationships', None)

    else: # set to defaults
        self.plot_net = p.plot_network
        self.net_font_name = 'DejaVu Sans'
        self.node_font_size = 16
        self.tit_font_size = 24
        self.net_layout = 'TB'
        self.edge_width = 2.0

        self.conc_node_font_color = 'Black'
        self.conc_node_color = 'LightCyan'
        self.conc_shape = 'ellipse'

        self.react_node_font_color = 'White'
        self.react_node_color = 'DarkSeaGreen'
        self.reaction_shape = 'rect'

        self.transp_node_font_color = 'White'
        self.transp_node_color = 'DarkSeaGreen'
        self.transporter_shape = 'diamond'

        self.chan_node_font_color = 'White'
        self.chan_node_color = 'DarkSeaGreen'
        self.channel_shape = 'pentagon'

        self.ed_node_font_color = 'White'
        self.ed_node_color = 'DarkSeaGreen'
        self.ed_shape = 'hexagon'

        self.vmem_node_font_color = 'White'
        self.vmem_node_color = 'maroon'
        self.vmem_shape = 'oval'

        self.subnets_dicts = None
        self.additional_edges = None


def plot_master_network(self, p):

    # Defer importation of optional runtime dependencies until necessary.
    pydot = libs.import_runtime_optional('pydot')

    # create a graph object
    base_graph = pydot.Dot(
        graph_type='digraph',
        concentrate=False,
        nodesep=0.1,
        ranksep=0.3,
        splines=True,
        strict=True,
        rankdir=self.net_layout,
    )

    # add Vmem to the graph
    nde = pydot.Node(
        'Vmem',
        style='filled',
        shape=self.vmem_shape,
        color=self.vmem_node_color,
        fontcolor=self.vmem_node_font_color,
        fontname=self.net_font_name,
        fontsize=self.node_font_size,
    )
    base_graph.add_node(nde)

    # add each substance as a node in the graph:
    for i, (name, val) in enumerate(self.cell_concs.items()):
        if name not in p.ions_dict:
            mol = self.molecules[name]

            nde = pydot.Node(
                name,
                style='filled',
                color=self.conc_node_color,
                shape=self.conc_shape,
                fontcolor=self.conc_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )
            base_graph.add_node(nde)

            if mol.simple_growth:
                # add node & edge for growth reaction component:
                rea_name = name + '_growth'
                rea_node = pydot.Node(
                    rea_name,
                    style='filled',
                    color=self.react_node_color,
                    shape=self.reaction_shape,
                    fontcolor=self.react_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )

                base_graph.add_node(rea_node)

                # if the substance has autocatalytic growth capacity add the edge in:
                base_graph.add_edge(pydot.Edge(
                    rea_name,
                    name,
                    arrowhead='normal',
                    coeff=1.0,
                    penwidth=self.edge_width,
                ))

                # add node & edge for decay reaction component:
                rea_name = name + '_decay'
                rea_node = pydot.Node(
                    rea_name,
                    style='filled',
                    color=self.react_node_color,
                    shape=self.reaction_shape,
                    fontcolor=self.react_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )

                base_graph.add_node(rea_node)

                # if the substance has autocatalytic decay capacity add the edge in:
                base_graph.add_edge(pydot.Edge(
                    name,
                    rea_name,
                    arrowhead='normal',
                    coeff=1.0,
                    penwidth=self.edge_width,
                ))

            if mol.ion_channel_gating:
                # if this substance gates for ion channels:

                # define a node corresponding to the ion channel:
                gated_node = pydot.Node(
                    mol.gating_channel_name,
                    style='filled',
                    color=self.chan_node_color,
                    shape=self.channel_shape,
                    fontcolor=self.chan_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(gated_node)

                # add the nodes for substance gating channel, if needed:
                if mol.gating_extracell:
                    substance_name = name + "_env"

                    nde = pydot.Node(
                        name,
                        style='filled',
                        color=self.conc_node_color,
                        shape=self.conc_shape,
                        fontcolor=self.conc_node_font_color,
                        fontname=self.net_font_name,
                        fontsize=self.node_font_size,
                    )
                    base_graph.add_node(nde)

                else:
                    substance_name = name

                base_graph.add_edge(pydot.Edge(
                    substance_name,
                    gated_node,
                    arrowhead='dot',
                    color='blue',
                    penwidth=self.edge_width,
                ))

                for ion_name in mol.gating_ion_name:
                    ion_Pmem = 'Pmem_' + ion_name

                    # add the edges for channel effect on ion concentration:
                    base_graph.add_edge(pydot.Edge(
                        gated_node,
                        ion_Pmem,
                        arrowhead='dot',
                        color='blue',
                        penwidth=self.edge_width,
                    ))

        else:  # add the ion as a node, but don't worry about growth/decay
            nde = pydot.Node(
                name,
                style='filled',
                color=self.conc_node_color,
                shape=self.conc_shape,
                fontcolor=self.conc_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )
            base_graph.add_node(nde)

        # add expression nodes for electrodiffusion/diffusion:
        # check the Dmem value of the substance:
        dmem_check = self.Dmem[name]

        # FIXME here is one place where Vmem can activate/inhibit the reaction itself...
        if dmem_check != 0.0:
            react_label = 'Pmem_'+ name # here, "Pmem_" stands for membrane permeability
            nde = pydot.Node(
                react_label,
                style='filled',
                color=self.ed_node_color,
                shape=self.ed_shape,
                fontcolor=self.ed_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )

            base_graph.add_node(nde)

            # add the environmental complement for this substance:
            name_env = name + '_env'
            nde = pydot.Node(
                name_env,
                style='filled',
                color=self.conc_node_color,
                shape=self.conc_shape,
                fontcolor=self.conc_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )

            base_graph.add_node(nde)

            base_graph.add_edge(pydot.Edge(
                name,
                react_label,
                arrowhead='normal',
                coeff=1.0,
                penwidth=self.edge_width,
            ))

            base_graph.add_edge(pydot.Edge(
                react_label,
                name_env,
                arrowhead='normal',
                coeff=1.0,
                penwidth=self.edge_width,
            ))

    # Deal with activators and inhibitors for substance growth.
    for i, name in enumerate(self.molecules):
        mol = self.molecules[name]
        # add regulatory as nodes in the graph:

        if mol.simple_growth:
            gname = name + '_growth'

            graph_influencers(
                self, base_graph, gname, mol.growth_activators_list,
                mol.growth_inhibitors_list, p, reaction_zone='cell',
                zone_tags_a = mol.growth_activators_zone,
                zone_tags_i = mol.growth_inhibitors_zone)

        if mol.ion_channel_gating:
            graph_influencers(
                self, base_graph, mol.gating_channel_name, mol.ion_activators_list,
                mol.ion_inhibitors_list, p, reaction_zone='cell',
                zone_tags_a = mol.ion_activators_zone,
                zone_tags_i = mol.ion_inhibitors_zone)

    #--------------------------------------------------------------------------
    # Basic Bioelectric relations

    # detail how Vmem is affected via membrane permeability relationships to core ions:
    base_graph.add_edge(pydot.Edge(
        'Pmem_Na', 'Vmem', arrowhead='dot', color='blue', penwidth=self.edge_width))
    base_graph.add_edge(pydot.Edge(
        'Pmem_K', 'Vmem', arrowhead='tee', color='red', penwidth=self.edge_width))
    base_graph.add_edge(pydot.Edge(
        'K_env', 'Vmem', arrowhead='dot', color='blue', penwidth=self.edge_width))

    #--------------------------------------------------------------------------

    # if there are any reactions in the cytosol, add them to the graph
    if len(self.reactions) > 0:
        for i, name in enumerate(self.reactions):
            nde = pydot.Node(
                name,
                style='filled',
                color=self.react_node_color,
                shape=self.reaction_shape,
                fontcolor=self.react_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )
            base_graph.add_node(nde)

    if len(self.reactions_env) > 0:
        for i, name in enumerate(self.reactions_env):
            nde = pydot.Node(
                name,
                style='filled',
                color=self.react_node_color,
                shape=self.reaction_shape,
                fontcolor=self.react_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )
            base_graph.add_node(nde)

    # If there are any reactions in the mitochondria, add them to the graph.
    if len(self.reactions_mit) > 0:
        for i, name in enumerate(self.reactions_mit):
            nde = pydot.Node(
                name,
                style='filled',
                color=self.react_node_color,
                shape=self.reaction_shape,
                fontcolor=self.react_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size)
            base_graph.add_node(nde)

    # If there are any env reactions, plot their edges on the graph.
    if len(self.reactions_env) > 0:
        for i, name in enumerate(self.reactions_env):
            rea = self.reactions_env[name]

            for i, react_name in enumerate(rea.reactants_list):
                rea_coeff = rea.reactants_coeff[i]
                react_name_e = react_name + '_env'

                nde = pydot.Node(
                    react_name_e,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )

                base_graph.add_node(nde)
                base_graph.add_edge(pydot.Edge(
                    react_name_e,
                    name,
                    arrowhead='normal',
                    coeff=rea_coeff,
                    penwidth=self.edge_width,
                ))

            for j, prod_name in enumerate(rea.products_list):
                prod_coeff = rea.products_coeff[j]
                prod_name_e = prod_name + '_env'

                nde = pydot.Node(
                    prod_name_e,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )

                base_graph.add_node(nde)
                base_graph.add_edge(pydot.Edge(
                    name,
                    prod_name_e,
                    arrowhead='normal',
                    coeff=prod_coeff,
                    penwidth=self.edge_width,
                ))

            graph_influencers(
                self, base_graph, name, rea.reaction_activators_list,
                rea.reaction_inhibitors_list, p, reaction_zone=rea.reaction_zone,
                zone_tags_a=rea.reaction_activators_zone,
                zone_tags_i=rea.reaction_inhibitors_zone)

    # If there are any reactions, plot their edges on the graph.
    if len(self.reactions) > 0:
        for i, name in enumerate(self.reactions):
            rea = self.reactions[name]

            for i, react_name in enumerate(rea.reactants_list):
                rea_coeff = rea.reactants_coeff[i]
                base_graph.add_edge(pydot.Edge(
                    react_name, name, arrowhead='normal', coeff=rea_coeff,
                    penwidth=self.edge_width))

            for j, prod_name in enumerate(rea.products_list):
                prod_coeff = rea.products_coeff[j]
                base_graph.add_edge(pydot.Edge(
                    name, prod_name, arrowhead='normal', coeff=prod_coeff,
                    penwidth=self.edge_width))

            graph_influencers(
                self, base_graph, name, rea.reaction_activators_list,
                rea.reaction_inhibitors_list, p, reaction_zone=rea.reaction_zone,
                zone_tags_a=rea.reaction_activators_zone,
                zone_tags_i=rea.reaction_inhibitors_zone)

    # If there are any mitochondria zone reactions, plot their edges on the
    # graph (and react/prod nodes).
    if len(self.reactions_mit) > 0:
        for i, name in enumerate(self.reactions_mit):
            rea = self.reactions_mit[name]

            for i, react_name in enumerate(rea.reactants_list):
                react_name += '_mit'

                nde = pydot.Node(
                    react_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size)

                base_graph.add_node(nde)

                rea_coeff = rea.reactants_coeff[i]
                base_graph.add_edge(pydot.Edge(
                    react_name, name, arrowhead='normal', coeff=rea_coeff,
                    penwidth=self.edge_width))

            for j, prod_name in enumerate(rea.products_list):
                prod_name += '_mit'

                nde = pydot.Node(
                    prod_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(nde)

                prod_coeff = rea.products_coeff[j]
                base_graph.add_edge(pydot.Edge(
                    name, prod_name, arrowhead='normal', coeff=prod_coeff,
                    penwidth=self.edge_width))

            graph_influencers(
                self, base_graph, name, rea.reaction_activators_list,
                rea.reaction_inhibitors_list, p, reaction_zone=rea.reaction_zone,
                zone_tags_a=rea.reaction_activators_zone,
                zone_tags_i=rea.reaction_inhibitors_zone)

    # If there are any transporters, plot them on the graph.
    if len(self.transporters) > 0:
        for name in self.transporters:
            nde = pydot.Node(
                name,
                style='filled',
                color=self.transp_node_color,
                shape=self.transporter_shape,
                fontcolor=self.transp_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )

            base_graph.add_node(nde)

        for name in self.transporters:
            trans = self.transporters[name]

            for i, (react_name, tag) in enumerate(zip(
                trans.reactants_list, trans.react_transport_tag)):
                rea_coeff = trans.reactants_coeff[i]

                if tag == 'cell_concs' or tag == 'mem_concs':
                    base_graph.add_edge(pydot.Edge(
                        react_name,
                        name,
                        arrowhead='normal',
                        coeff=rea_coeff,
                        penwidth=self.edge_width,
                    ))

                else:
                    if tag == 'env_concs':
                        react_name += '_env'

                    elif tag == 'mit_concs':
                        react_name += '_mit'

                    nde = pydot.Node(
                        react_name,
                        style='filled',
                        color=self.conc_node_color,
                        shape=self.conc_shape,
                        fontcolor=self.conc_node_font_color,
                        fontname=self.net_font_name,
                        fontsize=self.node_font_size,
                    )

                    base_graph.add_node(nde)
                    base_graph.add_edge(pydot.Edge(
                        react_name,
                        name,
                        arrowhead='normal',
                        coeff=rea_coeff,
                        penwidth=self.edge_width,
                    ))

            for j, (prod_name, tag) in enumerate(zip(
                trans.products_list, trans.prod_transport_tag)):
                prod_coeff = trans.products_coeff[j]

                if tag == 'cell_concs' or tag == 'mem_concs':
                    base_graph.add_edge(pydot.Edge(
                        name,
                        prod_name,
                        arrowhead='normal',
                        coeff=prod_coeff,
                        penwidth=self.edge_width,
                    ))

                else:
                    if tag == 'env_concs':
                        prod_name += '_env'

                    elif tag == 'mit_concs':
                        prod_name += '_mit'

                    nde = pydot.Node(
                        prod_name,
                        style='filled',
                        color=self.conc_node_color,
                        shape=self.conc_shape,
                        fontcolor=self.conc_node_font_color,
                        fontname=self.net_font_name,
                        fontsize=self.node_font_size,
                    )

                    base_graph.add_node(nde)
                    base_graph.add_edge(pydot.Edge(
                        name,
                        prod_name,
                        arrowhead='normal',
                        coeff=prod_coeff,
                        penwidth=self.edge_width,
                    ))

            graph_influencers(
                self, base_graph, name, trans.transporter_activators_list,
                trans.transporter_inhibitors_list, p, reaction_zone=trans.reaction_zone,
                zone_tags_a=trans.transporter_activators_zone,
                zone_tags_i=trans.transporter_inhibitors_zone)

    # If there are any channels, plot their type, ion  and Vmem relationships
    # in the graph.
    if len(self.channels) > 0:
        for i, name in enumerate(self.channels):
            chan = self.channels[name]
            chan_class = self.channels[name].channel_class
            channel_name = self.channels[name].channel_type

            #FIXME: Consider refactoring to use dictionary lookups. Turtle toes!
            if chan_class == 'Na':
                ion_name = ['Na']

            elif chan_class == 'K':
                ion_name = ['K']

            elif chan_class == 'Kir':
                ion_name = ['K']

            elif chan_class == 'Fun':
                ion_name = ['Na', 'K']

            elif chan_class == 'Ca':
                ion_name = ['Ca']

            elif chan_class == 'NaP':
                ion_name = ['Na']

            elif chan_class == 'Cat':
                ion_name = ['Na', 'K']

            elif chan_class == 'Cl':
                ion_name = ['Cl']

            else:
                ion_name = []

            # add the channel to the diagram:
            nde = pydot.Node(
                name,
                style='filled',
                color=self.chan_node_color,
                shape=self.channel_shape,
                fontcolor=self.chan_node_font_color,
                fontname=self.net_font_name,
                fontsize=self.node_font_size,
            )

            base_graph.add_node(nde)

            for ion_n in ion_name:
                ion_Pmem = 'Pmem_' + ion_n

                base_graph.add_edge(pydot.Edge(
                    name,
                    ion_Pmem,
                    arrowhead='dot',
                    color='blue',
                    penwidth=self.edge_width,
                ))

            graph_influencers(
                self, base_graph, name, chan.channel_activators_list,
                chan.channel_inhibitors_list, p, reaction_zone='cell',
                zone_tags_a=chan.channel_activators_zone,
                zone_tags_i=chan.channel_inhibitors_zone)

    # if there are any user-defined relations, add them in:

    if self.additional_edges is not None:
        for node_list in self.additional_edges:
            node1, node2, relation = node_list

            if not base_graph.get_node(node1):
                raise BetsePyDotException(
                    'Additional node "{}" not found.'.format(node1))

            if not base_graph.get_node(node2):
                raise BetsePyDotException(
                    'Additional node "{}" not found.'.format(node2))

            if relation == 'activation':
                arrowh = 'dot'
                arrowc = 'blue'

            elif relation == 'inhibition':
                arrowh = 'tee'
                arrowc = 'red'

            else:
                arrowh = 'normal'
                arrowc = 'black'

            base_graph.add_edge(pydot.Edge(
                node1, node2, arrowhead=arrowh, color=arrowc, penwidth=self.edge_width))

    # if sub-graphs are defined, re-jigger the network to have clusters/sub-networks:
    if (
        self.subnets_dicts is not None and
        self.subnets_dicts != 'None' and
        len(self.subnets_dicts) > 0
    ):
        base_graph = make_subgraphs(self, base_graph, p)

    return base_graph


def make_subgraphs(self, base_graph, p):

    # Defer importation of optional runtime dependencies until necessary.
    pydot, networkx = libs.import_runtime_optional('pydot', 'networkx')

    # Convert the pydot-version base_graph into a networkx function so we can
    # re-jigger nodes and make subgraphs.
    net_all = networkx.from_pydot(base_graph)

    subnetworks = {}

    master_graph = pydot.Dot(
        graph_type='digraph',
        concentrate=False,
        nodesep=0.1,
        ranksep=0.3,
        splines=True,
        strict=True,
        rankdir=self.net_layout,
    )

    # Begin by writing all nodes and edges from the main network to the new pydot network
    master_graph = write_nodes(net_all, master_graph)
    master_graph = write_edges(net_all, master_graph)

    for subnet_opts in self.subnets_dicts:
        subnet_name = subnet_opts['name']
        subnet_nodes = subnet_opts['nodes']

        # expand subnetwork nodes list with nearest neighbours from main graph:
        extended_node_list = []

        for nde in subnet_nodes:
            if nde in net_all.nodes():
                extended_node_list.append(nde)
                nn = net_all.neighbors(nde)

                for n in nn:
                    extended_node_list.append(n)

        # reassign subgraph nodes to the extended list:
        subnet_nodes = extended_node_list

        tit_font = subnet_opts['title font color']
        box_color = subnet_opts['box shading color']

        subnet = networkx.subgraph(net_all, subnet_nodes)
        subnetworks[subnet_name] = {
            'name': subnet_name,
            'nx subnet': subnet,
            'tit_font': tit_font,
            'box_color': box_color,
            'nodes': subnet_nodes,
        }

    for i, sn_dic in enumerate(subnetworks):
        cname = 'cluster_' + str(i)

        # declare subgraphs in pydot format:
        py_sub = pydot.Subgraph(
            cname,
            label=subnetworks[sn_dic]['name'],
            style='filled',
            fontcolor=subnetworks[sn_dic]['tit_font'],
            fontname=self.net_font_name,
            fontsize=self.tit_font_size,
            color=subnetworks[sn_dic]['box_color'],
        )

        subnetworks[sn_dic]['py subnet'] = py_sub

    # now each subnetwork has a networkx and empty pydot subgraph definition.
    # write each networkx subnet to the pydot subgraph:
    for i, sn_dic in enumerate(subnetworks):
        sn_py = write_nodes(
            subnetworks[sn_dic]['nx subnet'], subnetworks[sn_dic]['py subnet'])
        subnetworks[sn_dic]['py subnet'] = sn_py

        master_graph.add_subgraph(sn_py)
        master_graph = write_edges(subnetworks[sn_dic]['nx subnet'], master_graph)

    return master_graph


def graph_influencers(
    self,
    base_graph,
    name,
    a_list,
    i_list,
    p,
    reaction_zone='cell',
    zone_tags_a=None,
    zone_tags_i=None,
) -> None:
    '''
    Allow activators and inhibitors, which have various options, to be added to
    a graph.

    Parameters
    --------------
    graph:    The pydot graph object to add to
    name             Name of the main molecule being activated or inhibited
    a_list:   List of activator names
    i_list:   List of inhibitor names
    reaction_zone:  Zone for the main reaction
    zone_tags_a:     list of zones activator works from
    zone_tags_i:     list of zones inhibitor work from.

    Returns
    --------
    graph           Updated pydot graph object
    '''

    # Defer importation of optional runtime dependencies until necessary.
    pydot = libs.import_runtime_optional('pydot')

    if a_list != 'None' and a_list is not None:
        for act_name, zone_a in zip(a_list, zone_tags_a):
            if act_name.endswith('!'):  # if activator indicated to be independent
                # make the name in accordance with its actual identifier
                act_name = act_name[:-1]

            if reaction_zone == 'mit':
                act_name += '_mit'
                nde = pydot.Node(
                    act_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(nde)

            if zone_a == 'env':
                act_name = act_name + '_env'
                nde = pydot.Node(
                    act_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(nde)

            base_graph.add_edge(pydot.Edge(
                act_name, name, penwidth=self.edge_width, arrowhead='dot', color='blue'))

    if i_list != 'None' and i_list is not None:
        for inh_name, zone_i in zip(i_list, zone_tags_i):
            if reaction_zone == 'mit':
                inh_name += '_mit'
                nde = pydot.Node(
                    inh_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(nde)

            if zone_i == 'env':
                inh_name = inh_name + '_env'
                nde = pydot.Node(
                    inh_name,
                    style='filled',
                    color=self.conc_node_color,
                    shape=self.conc_shape,
                    fontcolor=self.conc_node_font_color,
                    fontname=self.net_font_name,
                    fontsize=self.node_font_size,
                )
                base_graph.add_node(nde)

            base_graph.add_edge(pydot.Edge(
                inh_name, name, arrowhead='tee', color='red', penwidth=self.edge_width))


def write_nodes(nx_graph, py_graph):

    # Defer importation of optional runtime dependencies until necessary.
    pydot = libs.import_runtime_optional('pydot')

    for nx_node_name in nx_graph.node:
        nx_dic = nx_graph.node[nx_node_name]

        clr = nx_dic.get('color', None)
        fntclr = nx_dic.get('fontcolor', None)
        fntnm = nx_dic.get('fontname', None)
        fntsz = nx_dic.get('fontsize', None)
        shp = nx_dic.get('shape', None)
        stl = nx_dic.get('style', None)

        # if (clr is not None and fntclr is not None and fntnm is not None and shp is not None and
        #             shp is not None and stl is not None):

        nde = pydot.Node(nx_node_name, color=clr, fontcolor=fntclr, fontname=fntnm, fontsize=fntsz,
                         shape=shp, style=stl)

        py_graph.add_node(nde)

        # else:
        #
        #     logs.log_warning("WARNING! You have requested a sub-graph node that does not exist!\n"
        #                      " Ignoring request!")

    return py_graph


def write_edges(nx_graph, py_graph):

    # Defer importation of optional runtime dependencies until necessary.
    pydot = libs.import_runtime_optional('pydot')

    for na, nb in nx_graph.edges():
        edge_dic = nx_graph.edge[na][nb]

        arrow = edge_dic.get('arrowhead', None)
        linew = edge_dic.get('penwidth', None)
        clr = edge_dic.get('color', 'black')

        # if arrow is not None and linew is not None and clr is not None:

        eg = pydot.Edge(na, nb, arrowhead=arrow, penwidth=linew, color=clr)
        py_graph.add_edge(eg)

        # else:
        #
        #     logs.log_warning("WARNING! You have requested a sub-graph edge that does not exist!\n"
        #                      " Ignoring request!")

    return py_graph



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
**Ionic enumerations** (i.e., :class:`enum.Enum` subclasses required to
differentiate between various types of ions).
'''

# ....................{ IMPORTS                           }....................
from betse.util.type import enums
from betse.util.type.types import GeneratorType

# ....................{ ITERATORS                         }....................
def iter_ion_names() -> GeneratorType:
    '''
    Generator yielding the unmodified name of each supported type of ion
    specified by the :class:`IonType` enumeration (in lexicographic order).

    Returns
    ----------
    GeneratorType
        Generator yielding each such name.
    '''

    # Let there be ion.
    yield from enums.iter_names(IonType)

# ....................{ ENUMS                             }....................
IonType = enums.make_enum(
    class_name='IonType',
    member_names=('Na', 'K', 'Ca', 'Cl', 'M', 'P',),
    is_ordered=True,
    doc='''
Ordered enumeration of all supported types of **ions** (i.e., atoms or
molecules having lost or gained electrons and hence net electric charge).

For brevity, the names of these types are intentionally condensed to their
standard abbreviations rather than elongated to their human-readable nouns
(e.g., ``Na`` rather than ``SODIUM``). Doing so also improves interoperability
with the remainder of the codebase, which first invented this nomenclature.

Attributes
----------
Na : enum
    Sodium cation Na+.
K : enum
    Potassium cation K+.
Ca : enum
    Calcium cation Ca2+.
Cl : enum
    Chloride anion Cl-.
M : enum
    Unidentified anion M-, synthetically manufactured to enforce
    charge-balancing across the environmental matrix for both simulation
    stability and correctness.
P : enum
    Anionic protein P-.
''')



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Apple macOS-specific facilities.
'''

# ....................{ IMPORTS                            }....................
import platform
from betse.exceptions import BetseOSException
from betse.util.io.log.logs import log_warning
from betse.util.type.decorator.decmemo import func_cached
from ctypes import CDLL, byref, c_int

# ....................{ EXCEPTIONS                         }....................
def die_unless_macos() -> None:
    '''
    Raise an exception unless the current platform is Apple macOS.

    See Also
    ----------
    :func:`is_macos`
        Further details.
    '''

    # Avoid circular import dependencies.
    from betse.util.os import oses

    # If the current platform is *NOT* macOS, raise an exception.
    if not is_macos():
        raise BetseOSException(f'{oses.get_name()} not macOS.')

# ....................{ TESTERS                            }....................
@func_cached
def is_macos() -> bool:
    '''
    ``True`` only if the current platform is Apple macOS, the operating
    system previously known as "OS X."
    '''

    return platform.system() == 'Darwin'


@func_cached
def is_aqua() -> bool:
    '''
    ``True`` only if the current process has access to the Aqua display server
    specific to macOS, implying this process to be headfull and hence support
    both CLIs and GUIs.

    See Also
    ----------
    https://developer.apple.com/library/content/technotes/tn2083/_index.html#//apple_ref/doc/uid/DTS10003794-CH1-SUBSECTION19
        "Security Context" subsection of "Technical Note TN2083: Daemons and
        Agents," a psuedo-human-readable discussion of the
        ``sessionHasGraphicAccess`` bit flag returned by the low-level
        ``SessionGetInfo()`` C function.
    '''

    # Avoid circular import dependencies.
    from betse.util.path.files import is_file
    from betse.util.os.command.cmdexit import SUCCESS

    # If the current platform is *NOT* macOS, return false.
    if not is_macos():
        return False
    # Else, the current platform is macOS.

    # If the system-wide Macho-O shared library providing the macOS
    # security context for the current process does *NOT* exist (after
    # following symbolic links)...
    if not is_file(_SECURITY_FRAMEWORK_DYLIB_FILENAME):
        # Emit a non-fatal warning. Theoretically, this shared library should
        # *ALWAYS* exist across all macOS versions (including those still
        # actively maintained as of 2023 Q2). Pragmatically, this shared library
        # appears to *NOT* exist (for unknown reasons) on GitHub Actions macOS
        # runners. He have no control over GitHub Actions. Let's complain! \o/
        log_warning(
            'macOS shared library "%s" not found.',
            _SECURITY_FRAMEWORK_DYLIB_FILENAME)

        # Return false.
        return False
    # Else, this shared library exists.

    # Attempt all of the following in a safe manner catching, logging, and
    # converting exceptions into a false return value. This tester is *NOT*
    # mission-critical and hence should *NOT* halt the application on
    # library-specific failures.
    try:
        # Dynamically load this library into the address space of this process.
        security_framework = CDLL(_SECURITY_FRAMEWORK_DYLIB_FILENAME)

        # Possibly non-unique identifier of the security session to request the
        # attributes of, signifying that of the current process.
        session_id = _SECURITY_SESSION_ID_CURRENT

        # Unique identifier of the requested security session, returned
        # by reference from the SessionGetInfo() C function called below. This
        # identifier is useless for our purposes and hence ignored below.
        session_id_real = c_int(0)

        # Attributes bit field of the requested security session, returned by
        # reference from the SessionGetInfo() C function called below.
        session_attributes = c_int(0)

        # C-style error integer returned by calling the SessionGetInfo() C
        # function exported by this Macho-O shared library, passing:
        #
        # * The input non-unique session identifier by value.
        # * The output unique session identifier by reference.
        # * The output session attributes integer by reference.
        session_errno = security_framework.SessionGetInfo(
            session_id, byref(session_id_real), byref(session_attributes))

        # This process has access to the Aqua display server if and only if...
        return (
            # The above function call succeeded *AND*...
            session_errno == SUCCESS and
            # The session attributes bit field returned by this call has the
            # corresponding bit flag enabled.
            bool(
                session_attributes.value & _SECURITY_SESSION_HAS_GRAPHIC_ACCESS)
        )
    # If the above logic failed with any exception...
    except Exception as exception:
        # Human-readable exception message harvested from this exception,
        # defined as either:
        # * If this exception is a platform-specific "OSError" (as is likely due
        #   to calling low-level platform-specific macOS kernel functions
        #   above), the "OSError.strerror" instance variable of this exception.
        # * Else, the standard string representation of this exception.
        exception_message = getattr(exception, 'strerror', str(exception))

        # Log a non-fatal warning informing users of this failure.
        log_warning(
            'macOS SessionGetInfo() function failed, as %s.', exception_message)

    # Assume this process to *NOT* have access to the Aqua display server.
    return False

# ....................{ PRIVATE ~ constants                }....................
_SECURITY_FRAMEWORK_DYLIB_FILENAME = (
    '/System/Library/Frameworks/Security.framework/Security')
'''
Absolute path of the system-wide ``"Security.framework"`` Macho-O shared library
providing the macOS-specific security context for the current process.

This library is dynamically loadable into the address space of the current
process with the :class:`ctypes.CDLL` class. Since all Macho-O shared libraries
necessarily have the filetype ``".dylib"``, this filetype can be safely omitted.
'''


_SECURITY_SESSION_ID_CURRENT = -1
'''
Magic integer defined as `callerSecuritySession` by the macOS-specific
`/System/Library/Frameworks/Security.Framework/Headers/AuthSession.h` C header
suitable for passing to C functions accepting parameters of C type
`SecuritySessionId` (e.g., `SessionGetInfo()`).

When passed, this integer signifies the **current security session** (i.e., the
the security session to which the current process belongs).

See Also
----------
https://opensource.apple.com/source/libsecurity_authorization/libsecurity_authorization-32564/lib/AuthSession.h
    C header defining this magic integer.
'''


_SECURITY_SESSION_HAS_GRAPHIC_ACCESS = 0x0010
'''
Bit flag defined as `sessionHasGraphicAccess` by the macOS-specific
`/System/Library/Frameworks/Security.Framework/Headers/AuthSession.h` C header
masking the attributes bit field returned by the `SessionGetInfo()` C function
also declared by that header.

When enabled, this bit signifies the current process to have access to the Aqua
display server and hence be headfull (rather than headless).

See Also
----------
https://opensource.apple.com/source/libsecurity_authorization/libsecurity_authorization-32564/lib/AuthSession.h
    C header defining this bit flag.
'''



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Class hierarchy collectively implementing various methods for assigning a subset
of the total cell population to the corresponding tissue profile.
'''

# ....................{ IMPORTS                            }....................
import numpy as np
from betse.exceptions import BetseImageException
from betse.lib.pil import pilnumpy
from betse.lib.pil.pilnumpy import ImageModeType
from betse.science.tissue.picker.tispickcls import TissuePickerABC
from betse.util.math.mathinterp import interp2d_linear
from betse.util.path import files, pathnames
from betse.util.type.types import type_check, NumericSimpleTypes, SequenceTypes
from numpy import ndarray
# from scipy import interpolate
from scipy.interpolate import RectBivariateSpline
from scipy.spatial import ConvexHull

# ....................{ CLASSES ~ utility                  }....................
#FIXME: Generalize this class to support images of arbitrary (possibly
#non-square) dimensions.
#FIXME: Document all undocumented attributes.
class TissuePickerImageMask(object):
    '''
    Cell profile-specific **image mask** (i.e., low-level object converting an
    well-formatted image into multiple SciPy-based interpolation functions
    clipping passed points to that image's colored pixel area).

    Design
    ----------
    This utility class is principally intended to be instantiated by the
    :class:`TissuePickerImage` class, but can (in theory) be instantiated
    directly by sufficiently careful external callers.

    This utility class is typically large in terms of memory consumption. Hence,
    the :class:`TissuePickerImage` class only locally instantiates instances of
    this class for the duration of method calls rather than permanently
    persisting such instances as instance variables of that class.

    Attributes
    ----------
    clipping_matrix : ndarray
        Numpy matrix defining this threshholded image.
    clipping_function : func
        SciPy-based interpolation function accepting an ``(x, y)`` point and
        returning ``1.0`` if that point resides outside this bitmap's colored
        pixel area or ``0.0`` otherwise. This function permits callers to filter
        a passed set of points in the space defined by :attr:`p.wsx` for the
        subset residing within this area.
    clipping_function_fast : func
        Fast variant of :attr:`clipping_function`, but otherwise sharing the
        same API.
    msize : int
        Size in pixels of each square dimension of this image, equivalent to
        both the width and height of this image.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(
        self,
        filename: str,
        x_min: NumericSimpleTypes,
        x_max: NumericSimpleTypes,
        y_min: NumericSimpleTypes,
        y_max: NumericSimpleTypes,
    ) -> None:
        '''
        Load, initialize, and create a threshholded interpolation matrix from
        the passed image mask file.

        Format
        ----------
        This image should ideally be completely threshholded. Each pure black
        pixel (i.e., with red, green, and blue color components all 0) of this
        image collectively defines either:

        * The area to be used as the clipping mask for the cell cluster.
        * The area for a tissue, cut, or boundary profile.

        This image *must* additionally:

        * Be square (i.e., have equal pixel dimensions). For consistency, image
          dimensions of 500x500 pixels are recommended.
        * Contain *no* alpha transparency layer, a SciPy-based constraint and
          hence *not* amenable to change.

        Parameters
        ----------
        filename : str
            Absolute or relative filename of this image.
        x_min : NumericSimpleTypes
            Minimum X coordinate accepted by these interpolation functions.
        x_max : NumericSimpleTypes
            Maximum X coordinate accepted by these interpolation functions.
        y_min : NumericSimpleTypes
            Minimum Y coordinate accepted by these interpolation functions.
        y_max : NumericSimpleTypes
            Maximum Y coordinate accepted by these interpolation functions.

        Raises
        ----------
        BetseFileException
            If this image does *not* exist.
        BetseImageeException
            If this image has either:
            * Non-square dimensions (i.e., differing width and height).
            * An alpha transparency layer.
        '''

        # Load this bitmap as a grayscale Numpy array, preserving floating point
        # precision across this necessarily lossy reduction.
        bitmap = pilnumpy.load_image(
            filename=filename, mode=ImageModeType.GRAYSCALE_FLOAT)

        # If this bitmap has non-square dimensions, raise an exception.
        if bitmap.shape[0] != bitmap.shape[1]:
            raise BetseImageException(
                'Image "{}" dimensions non-square '
                '(i.e., width {} != height {}).'.format(
                    filename, bitmap.shape[0], bitmap.shape[1]))

        # Find the black pixels. (This is a really basic threshholding!)
        point_inds = (bitmap != 255).nonzero()

        # New matrix the same shape as the image and set values to 0 or 1.
        self.msize = bitmap.shape[0]
        self.clipping_matrix = np.zeros((self.msize, self.msize))
        self.clipping_matrix[point_inds] = 1.0
        self.clipping_matrix = np.flipud(self.clipping_matrix)

        # Create spatial data vectors that span the extent of the cell seeds and
        # match bitmap pixel number.
        xpts = np.linspace(x_min, x_max, self.msize)
        ypts = np.linspace(y_min, y_max, self.msize)

        #FIXME: This doesn't appear to be quite right. Either that, or our
        #interp2d_linear() implementation isn't quite right. Gah!
        # Unravel these vectors as required by the interp2d_linear() function.
        xpts_2d_grid, ypts_2d_grid = np.meshgrid(xpts, ypts)
        xpts_2d = xpts_2d_grid.ravel()
        ypts_2d = ypts_2d_grid.ravel()
        clipping_matrix_2d = self.clipping_matrix.ravel()

        # Create an interpolation function that returns zero if the query point
        # is outside the mask and 1 if the query point is in the mask.
        self.clipping_function = interp2d_linear(
            xpts, ypts, self.clipping_matrix)
            # xpts_2d, ypts_2d, clipping_matrix_2d)
        # self.clipping_function = interpolate.interp2d(
        #     xpts, ypts, self.clipping_matrix)
        self.clipping_function_fast = RectBivariateSpline(
            xpts, ypts, self.clipping_matrix)

        # Store some additional information relating to bounding polygon of the
        # clipping image.
        Xclip, Yclip = np.meshgrid(xpts, ypts)

        indsN = (self.clipping_matrix == 1.0).nonzero()

        ptsx = Xclip[indsN].ravel()
        ptsy = Yclip[indsN].ravel()

        # points stack
        ppts = np.column_stack((ptsx, ptsy))

        # Calculate convex hull of shape:
        bclip = ConvexHull(ppts)

        # calculate
        bverts = bclip.vertices

        # store boundary points of the clipping poly representing the image shape:
        bx = ptsx[bverts]
        by = ptsy[bverts]

        # store the points of the clipping poly curve:
        self.clipcurve = np.column_stack((bx, by))

    # ..................{ GETTERS                            }..................
    @type_check
    def get_clipped_points_index(
        self,
        points_x: SequenceTypes,
        points_y: SequenceTypes,
    ) -> ndarray:
        '''
        One-dimensional Numpy array of the indices of all passed points that
        reside inside and are thus clipped by this image mask's colored area.

        Parameters
        -----------
        points_x : SequenceTypes
            Sequence of the X coordinates of all points to be clipped.
        points_y : SequenceTypes
            Sequence of the Y coordinates of all points to be clipped.

        Returns
        -----------
        ndarray
            One-dimensional Numpy array of the indices of all clipped points.
        '''

        # List of the indices of all passed points clipped by this method.
        clipped_points_index = []

        # For the index and X and Y coordinate of each passed point...
        for points_index, (x, y) in enumerate(zip(points_x, points_y)):
            # If this point resides inside this image mask, clip this point.
            if self.clipping_function(x,y) != 0.0:
                clipped_points_index.append(points_index)

        # Coerce this possibly empty list into a guaranteed integer Numpy array.
        # Since Numpy has no means of deciding whether an empty list should be
        # coerced into an integer or floating point Numpy array, Numpy sensibly
        # defaults to the latter functionality. While that's typically what one
        # wants, that is *NOT* what callers expect in this case. For caller
        # sanity, Numpy *MUST* be instructed to produce an integer array.
        return np.asarray(clipped_points_index, dtype=int)

# ....................{ CLASSES ~ picker                   }....................
class TissuePickerImage(TissuePickerABC):
    '''
    Image-based tissue picker, matching all cells residing inside the colored
    pixel area defined by an associated on-disk image mask file.

    Attributes
    ----------
    filename : str
        Absolute path of this image mask.
    '''

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self, filename: str, dirname: str) -> None:
        '''
        Initialize this tissue picker.

        Parameters
        ----------
        filename : str
            Absolute or relative filename of this image mask. If relative (i.e.,
            *not* prefixed by a directory separator), this filename will be
            canonicalized into an absolute filename relative to the passed
            absolute dirname.
        dirname : str
            Absolute dirname of the directory containing this image mask. If the
            ``filename`` parameter is relative, that filename will be prefixed
            by this path to convert that filename into an absolute path; else,
            this dirname is ignored.
        '''

        # If this is a relative path, convert this into an absolute path
        # relative to the directory containing the source configuration file.
        if pathnames.is_relative(filename):
            filename = pathnames.join(dirname, filename)

        # If this absolute path is *NOT* an existing file, raise an exception.
        files.die_unless_file(filename)

        # Classify this parameter.
        self.filename = filename

    # ..................{ PICKERS                            }..................
    @type_check
    def pick_cells(
        self,
        cells: 'betse.science.cells.Cells',
        p:     'betse.science.parameters.Parameters',
    ) -> SequenceTypes:

        # Cell profile-specific image mask.
        image_mask = self.get_image_mask(cells)

        # One-dimensional Numpy array of the indices of all cells whose centres
        # reside inside and are thus clipped by this image mask's colored area.
        clipped_points_index = image_mask.get_clipped_points_index(
            points_x=cells.cell_centres[:,0],
            points_y=cells.cell_centres[:,1])

        # Return this array.
        return clipped_points_index

    # ..................{ GETTERS                            }..................
    @type_check
    def get_image_mask(
        self, cells: 'betse.science.cells.Cells') -> TissuePickerImageMask:
        '''
        Cell profile-specific image mask, abstracting the low-level pixels of
        this image file into higher-level Python objects.

        Parameters
        ----------
        cells : Cells
            Current cell cluster.

        Returns
        ----------
        TissuePickerImageMask
            Cell profile-specific image mask implementing this picker.
        '''

        return TissuePickerImageMask(
            filename=self.filename,
            x_min=cells.xmin,
            x_max=cells.xmax,
            y_min=cells.ymin,
            y_max=cells.ymax,
        )



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :mod`betse.util.type.numeric.floats` submodule.
'''

# ....................{ IMPORTS                           }....................
# import pytest

# ....................{ TESTS ~ base 10                   }....................
def test_float_get_base_10_exponent() -> None:
    '''
    Unit test the :func:`betse.util.type.numeric.floats.get_base_10_exponent`
    getter.
    '''

    # Defer heavyweight imports.
    from betse.util.type.numeric import floats

    # Assert this getter to behave as expected on common edge cases.
    assert floats.get_base_10_exponent(7.77e+66) == 66
    assert floats.get_base_10_exponent(6.66e-77) == -77
    assert floats.get_base_10_exponent(0.000123456789) == -4
    assert floats.get_base_10_exponent(100001234567.9) == 11


def test_float_get_base_10_precision() -> None:
    '''
    Unit test the :func:`betse.util.type.numeric.floats.get_base_10_precision`
    getter.
    '''

    # Defer heavyweight imports.
    from betse.util.type.numeric import floats

    # Assert this getter to behave as expected on common edge cases, including:
    #
    # * A large float formatted in scientific notation by the str() builtin.
    # * A small float formatted in scientific notation by the str() builtin.
    # * A small float with a leading digit formatted in decimal notation by the
    #   str() builtin.
    # * A small float with no leading digit formatted in decimal notation by the
    #   str() builtin.
    #
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    # CAUTION: Python implicitly reformats all floating point numbers smaller
    # than 0.0001 when converted to strings in scientific rather than decimal
    # notation. Hence, small floats should be no smaller than 0.0001.
    #!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    assert floats.get_base_10_precision(7.77e+66) == 66
    assert floats.get_base_10_precision(6.66e-77) == 77
    assert floats.get_base_10_precision(0.000123456789) == 12
    assert floats.get_base_10_precision( .000123456789) == 12



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **Python implementation** (i.e., Python interpreter conforming to the
established Python syntax and semantics of the CPython reference
implementation) facilities.

Caveats
----------
Implementation-specific functions are generally considered poor form. Call
these functions *only* where necessary.
'''

# ....................{ IMPORTS                           }....................
import platform
from betse.util.type.decorator.decmemo import func_cached
from betse.util.type.iterable.mapping.mapcls import OrderedArgsDict

# ....................{ TESTERS                           }....................
@func_cached
def is_cpython() -> bool:
    '''
    ``True`` only if the active Python interpreter is an instance of the
    official CPython implementation.
    '''

    # Depressingly, this actually appears to be the most efficacious test.
    return get_name() == 'CPython'


@func_cached
def is_pypy() -> bool:
    '''
    ``True`` only if the active Python interpreter is an instance of the
    third-party PyPy implementation.
    '''

    # Depressingly, this actually appears to be the most efficacious test.
    return get_name() == 'PyPy'

# ....................{ GETTERS                           }....................
@func_cached
def get_name() -> str:
    '''
    Human-readable name of the active Python interpreter's implementation
    (e.g., ``CPython``, ``IronPython``, ``Jython``, ``PyPy``).
    '''

    return platform.python_implementation()

# ....................{ GETTERS ~ metadata                }....................
def get_metadata() -> OrderedArgsDict:
    '''
    Ordered dictionary synopsizing the active Python interpreter's
    implementation.

    This function aggregates the metadata reported by the reasonably
    cross-platform module `platform` into a simple dictionary.
    '''

    # This dictionary.
    metadata = OrderedArgsDict(
        'name', get_name(),
        'vcs revision', platform.python_revision() or 'none',
        'vcs branch', platform.python_branch() or 'none',
        'compiler', platform.python_compiler(),
    )

    # 2-tuple providing this interpreter's build number and date as strings.
    python_build = platform.python_build()

    # Append this metadata.
    metadata['build number'] = python_build[0]
    metadata['build data'] = python_build[1]

    # Return this dictionary.
    return metadata



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Unit tests exercising the :mod:`betse.util.type.iterable.itertest` submodule.
'''

# ....................{ IMPORTS                           }....................

# ....................{ TESTS ~ default                   }....................
#FIXME: Exercise the additional edge-case of a non-empty non-set of an
#arbitrary number of items, at least one of which (ideally the last) is
#non-hashable. Which begs the question, of course: what's trivially
#non-hashable?
def test_is_items_unique() -> None:
    '''
    Unit test the
    :func:`betse.util.type.iterable.itertest.is_items_unique` function.
    '''

    # Defer heavyweight imports.
    from betse.util.type.iterable import generators, itertest

    # Short iterable containing only unique items.
    mysidian_legend_unique = (
        'One to be born',
        'From a dragon',
        'Hoisting the light',
        'And the dark',
        'Arises high up',
        'In the sky to',
        'The still land.',
        'Veiling the moon with',
        'The Light of Eternity,',
        'It brings',
        'Another promise',
        'To Mother Earth with',
        'A bounty and mercy..',
    )

    # Short iterable containing one or more duplicate items.
    mysidian_legend_nonunique = (
        mysidian_legend_unique + (mysidian_legend_unique[-1],))

    # Long iterable containing only unique items.
    burning_legend_unique = tuple(range(256))

    # Long iterable containing one or more duplicate items.
    burning_legend_nonunique = (
        burning_legend_unique + (burning_legend_unique[-1],))

    # Exercise the edge case when this iterable is an empty generator.
    assert itertest.is_items_unique(iterable=generators.empty_generator())

    # Exercise the edge case when this iterable is an empty non-generator.
    assert itertest.is_items_unique(iterable=())

    # Exercise the edge case when this iterable is a non-empty set.
    assert itertest.is_items_unique(iterable=set(mysidian_legend_unique))

    # Exercise the edge case when this iterable is a non-empty non-set of less
    # than 64 hashable unique items.
    assert itertest.is_items_unique(iterable=mysidian_legend_unique)

    # Exercise the edge case when this iterable is a non-empty non-set of less
    # than 64 hashable non-unique items.
    assert (
        itertest.is_items_unique(iterable=mysidian_legend_nonunique) is False)

    # Exercise the edge case when this iterable is a non-empty non-set of
    # greater than or equal to than 64 hashable unique items.
    assert itertest.is_items_unique(iterable=burning_legend_unique)

    # Exercise the edge case when this iterable is a non-empty non-set of
    # greater than or equal to than 64 hashable non-unique items.
    assert (
        itertest.is_items_unique(iterable=burning_legend_nonunique) is False)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Subsidiary entry point of this application's command line interface (CLI),
preserving backward compatibility with prior versions.
'''

# ....................{ IMPORTS                           }....................
from betse.__main__ import main
from betse.util.os.command import cmdexit

# ....................{ MAIN                              }....................
# If this module is imported from the command line, run this application's CLI;
# else, noop. For POSIX compliance, the exit status returned by this function
# is propagated to the caller as this script's exit status.
if __name__ == '__main__':
    cmdexit.exit_with_status(main())



#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
CLI-specific functional tests exercising all miscellaneous subcommands (e.g.,
`betse info`), thus excluding those specific to networks and simulations.
'''

# ....................{ IMPORTS                            }....................
import pytest
from betse.util.test.pytest.mark.pytskip import (
    skip_unless_lib_runtime_optional)

# ....................{ TESTS                              }....................
def test_cli_no_arg(betse_cli: 'CLITester') -> None:
    '''
    Test the ``betse`` command passed no arguments.

    Parameters
    ----------
    betse_cli : CLITester
        Object encapsulating the BETSE CLI.
    '''

    betse_cli.run()


def test_cli_info(betse_cli: 'CLITester') -> None:
    '''
    Test the ``betse info`` subcommand.

    Parameters
    ----------
    betse_cli : CLITester
        Object encapsulating the BETSE CLI.
    '''

    betse_cli.run('info')


def test_cli_config(
    betse_cli: 'CLITester',
    betse_temp_dir: 'py._path.local.LocalPath',
) -> None:
    '''
    Test the ``betse config`` subcommand, creating a configuration in a
    temporary directory specific to this test.

    While other tests requesting any fixture whose name is prefixed by
    ``betse_sim_config_`` (e.g., the ``betse_sim_config_default`` fixture)
    already create such configurations, they do so by directly calling a
    comparatively low-level function of the BETSE API rather than by running a
    comparatively high-level subcommand of the BETSE CLI. Since these two logic
    paths are different (albeit related), this test ensures the latter to be
    exercised.

    Parameters
    ----------
    betse_cli : CLITester
        Object encapsulating the BETSE CLI.
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to the current
        test.
    '''

    # Absolute path of this configuration file in this temporary directory.
    sim_config_filepath = betse_temp_dir.join('sim_config.yaml')

    # Create this file. (Avoid shell-quoting this path. Doing so unnecessarily
    # adds an additional level of quoting... which is bad.)
    betse_cli.run('config', str(sim_config_filepath))

    # Assert this file to have been created.
    assert sim_config_filepath.check(file=1)

# ....................{ TESTS ~ parametrized               }....................
@pytest.mark.parametrize(
    ('profile_type',), (
        # Profile types leveraging stdlib functionality guaranteed to exist.
        ('none',),
        ('call',),

        # Profile types leveraging third-party packages necessitating checking.
        pytest.param(
            'size', marks=skip_unless_lib_runtime_optional('pympler')),
    ),
)
def test_cli_profile(
    betse_cli: 'CLITester',
    betse_temp_dir: 'py._path.local.LocalPath',
    profile_type: str,
) -> None:
    '''
    Test profiling of an arbitrary ``betse`` command with the parametrized
    profiling type, serialized to a temporary file specific to this test.

    Parameters
    ----------
    betse_cli : CLITester
        Object encapsulating the BETSE CLI.
    betse_temp_dir : py._path.local.LocalPath
        Object encapsulating a temporary directory isolated to the current
        test.
    profile_type : str
        Type of profiling to perform -- either:

        * ``none``, performing no profiling.
        * ``call``, performing call-granularity profiling.
        * ``size``, performing memory profiling.
    '''
        # * `line`, performing line-granularity profiling.

    # CLI option enabling this type of profiling.
    profile_type_option = '--profile-type={}'.format(profile_type)

    # If this is a profiler outputing no profile file...
    if profile_type in ('none', 'size',):
        # Profile the basic "betse" command to only the logfile. While the
        # absolute path of a profile file could also be passed, doing so
        # requires inefficiently creating a temporary directory. Since these
        # profilers output no profile file, this directory would remain empty.
        betse_cli.run(profile_type_option)
    else:
        # Absolute path of the profile file isolated to a test-specific
        # temporary directory output by the following command.
        profile_filepath = betse_temp_dir.join('profile.prof')

        # Profile the basic "betse" command to the logfile *AND* this file.
        betse_cli.run(
            profile_type_option, '--profile-file={}'.format(profile_filepath))

        # Assert this file to have been created.
        assert profile_filepath.check(file=1)



#!/usr/bin/env python3
# --------------------( LICENSE                           )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
Low-level **units** (i.e., standard for quantitative measurement of observable
quantities) facilities.
'''

# ....................{ IMPORTS                           }....................
from betse.lib.numpy import nparray
from betse.util.type import types
# from betse.util.type.numeric import ints
from betse.util.type.types import (
    type_check, NumericSimpleTypes, NumericOrIterableTypes,)

# ....................{ CONSTANTS ~ inverse               }....................
INVERSE_CENTI = 1e2
'''
Inverse of the centi- unit prefix (i.e., ``10**−2``), commonly used as a
multiplicative factor for upscaling quantities from centi-prefixed units to
unprefixed units.
'''


INVERSE_MILLI = 1e3
'''
Inverse of the milli- unit prefix (i.e., ``10**−3``), commonly used as a
multiplicative factor for upscaling quantities from milli-prefixed units to
unprefixed units.
'''


INVERSE_MICRO = 1e6
'''
Inverse of the micro- unit prefix (i.e., ``10**−6``), commonly used as a
multiplicative factor for upscaling quantities from micro-prefixed units to
unprefixed units.

Specifically, this is equivalent to the multiplicative factor for upscaling
cell cluster coordinates from micro-prefixed units to unprefixed units (e.g.,
to improve the readability of these coordinates in user-friendly exports).
'''

# ....................{ UPSCALERS ~ cell : data           }....................
@type_check
def upscale_units_centi(
    data: NumericOrIterableTypes) -> NumericOrIterableTypes:
    '''
    Upscale the contents of the passed number or iterable of numbers whose
    units are assumed to be centi-prefixed (i.e., factors of ``10**-2``).

    This function does *not* modify the passed object. If this object is:

    * A scalar number (e.g., :class:`int`, :class:`float`), this function
      returns a new scalar number of the same type multiplied by a positive
      constant.
    * A non-Numpy iterable (e.g., :class:`list`), this function returns a new
      Numpy array first converted from this iterable and then multiplied by a
      positive constant.
    * A Numpy array, this function returns a new Numpy array first copied from
      the passed array and then multiplied by a positive constant.

    Parameters
    ----------
    data : NumericOrIterableTypes
        Either a number of iterable of numbers to upscale.

    Returns
    ----------
    object
        Upscaled object as described above.
    '''

    return _upscale_data_in_units(data=data, factor=INVERSE_CENTI)


@type_check
def upscale_units_milli(
    data: NumericOrIterableTypes) -> NumericOrIterableTypes:
    '''
    Upscale the contents of the passed number or iterable of numbers whose
    units are assumed to be milli-prefixed (i.e., factors of ``10**-3``).

    See Also
    ----------
    :func:`upscale_units_centi`
        Further details.
    '''

    return _upscale_data_in_units(data=data, factor=INVERSE_MILLI)


@type_check
def upscale_units_micro(
    data: NumericOrIterableTypes) -> NumericOrIterableTypes:
    '''
    Upscale the contents of the passed number or iterable of numbers whose
    units are assumed to be micro-prefixed (i.e., factors of ``10**-6``).

    See Also
    ----------
    :func:`upscale_units_centi`
        Further details.
    '''

    return _upscale_data_in_units(data=data, factor=INVERSE_MICRO)

# ....................{ UPSCALERS ~ cell : coordinates    }....................
@type_check
def upscale_coordinates(
    coordinates: NumericOrIterableTypes) -> NumericOrIterableTypes:
    '''
    Upscale the contents of the passed number or sequence of numbers whose
    units are assumed to be denominated in micrometers (i.e., ``10**-6``),
    typically for converting spatial data into X and Y coordinates suitable for
    plot and animation display.

    Parameters
    ----------
    coordinates : NumericOrIterableTypes
        Number or sequence to be upscaled.

    Returns
    ----------
    NumericOrIterableTypes
        Either:

        * If this data is numeric, this number upscaled by this multiplier.
        * If this data is sequential, this sequence converted into a Numpy
          array whose elements are upscaled by this multiplier.

    See Also
    ----------
    :func:`upscale_units_milli`
        Further details.
    '''

    return upscale_units_micro(coordinates)


@type_check
def upscale_coordinates_tuple(
    *cells_coordinates: NumericOrIterableTypes) -> tuple:
    '''
    Upscale the contents of each passed number or sequence of numbers whose
    units are assumed to be denominated in micrometers (i.e., ``10**-6``),
    returning an n-tuple of each upscaled number or sequence of numbers.

    Parameters
    ----------
    cells_coordinates : tuple[NumericOrIterableTypes]
        Tuple of all numbers and sequences to be upscaled.

    Returns
    ----------
    tuple
        N-tuple for ``N`` the number of passed positional arguments such that
        each element is the corresponding positional argument upscaled by the
        :func:`upscale_coordinates` function.
    '''

    return tuple(
        upscale_coordinates(cell_coordinates)
        for cell_coordinates in cells_coordinates
    )

# ....................{ UPSCALERS ~ private               }....................
@type_check
def _upscale_data_in_units(
    data: NumericOrIterableTypes, factor: NumericSimpleTypes) -> (
    NumericOrIterableTypes):
    '''
    Upscale the contents of the passed object by the passed multiplier,
    typically under the assumption that these contents are denominated in the
    reciprocal units of this multiplier (i.e., ``10**6`` for micrometers).

    Parameters
    ----------
    data : NumericOrIterableTypes
        Number or sequence to be upscaled.
    factor : NumericSimpleTypes
        Reciprocal of the units this object is denominated in.

    Returns
    ----------
    NumericOrIterableTypes
        Either:

        * If this data is numeric, this number upscaled by this multiplier.
        * If this data is sequential, this sequence converted into a Numpy
          array whose elements are upscaled by this multiplier.

    See Also
    ----------
    :func:`upscale_units_milli`
        Further details.
    '''

    # If the passed object is numeric, return this number upscaled.
    if types.is_numeric(data):
        return factor * data
    # Else, this object is a sequence. Return this sequence converted into a
    # Numpy array and then upscaled.
    else:
        return factor * nparray.from_iterable(data)



#!/usr/bin/env python3
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

import numpy as np
from betse.science import sim_toolbox as stb


class MoveChannel(object):

    """
    Movement of ion pumps and channels to potentially create directional fluxes in individual cells.

    This is presently simulated by calculating the Nernst-Planck concentration flux of a weighting
    agent, rho, which moves under its own concentration gradient (a homogeneity restoring influence)
    as well as under the influence of the extracellular voltage gradient and fluid flows tangential
    to the membrane.

    """

    def __init__(self, sim, cells, p):

        sim.rho_pump = np.ones(sim.mdl)
        sim.rho_channel = np.ones(sim.mdl)

        # transfer rate for entities in the membrane:
        # self.alpha = p.D_membrane/(p.tm*p.cell_height*cells.num_mems[cells.mem_to_cells])
        #
        # # initialize pumps:
        # self.pgrad_x = np.zeros(sim.mdl)
        # self.pgrad_y = np.zeros(sim.mdl)
        #
        # # initialize channels:
        # self.cgrad_x = np.zeros(sim.mdl)
        # self.cgrad_y = np.zeros(sim.mdl)

    def run(self, sim, cells, p):

        self.update_channel(sim, cells, p)
        self.update_pump(sim, cells, p)

    def update_pump(self, sim, cells, p):

        cav = 1.0  # concentration at cell centre
        cpi = sim.rho_pump  # concentration at membrane
        z = p.z_pump  # charge of ion
        Do = p.D_membrane  # diffusion constant of ion

        cap = (cav + cpi) / 2  # concentration at midpoint between cell centre and membrane
        cgp = (cpi - cav) / cells.R_rads  # concentration gradients

        # Nernst-Planck equation to calculate flux:
        cfluxpo = -Do * cgp + ((Do * p.q * cap * z) / (p.kb * sim.T)) * sim.Ec

        # as no net mass must leave this intracellular movement, make the flux divergence-free:
        cfluxp = stb.single_cell_div_free(cfluxpo, cells)

        # calculate the actual concentration at membranes by unpacking to concentration vectors:
        sim.rho_pump = cpi + cfluxp * (cells.mem_sa / cells.mem_vol) * p.dt

        # calculate the equillibrium gradient vector pointing from maximum to minimum density on the membrane:
        # ceqm_x = ((p.z_pump * p.q) / (p.kb * p.T)) * 1.0 * sim.J_cell_x[cells.mem_to_cells] * (1/sim.sigma)
        # ceqm_y = ((p.z_pump * p.q) / (p.kb * p.T)) * 1.0 * sim.J_cell_y[cells.mem_to_cells] * (1/sim.sigma)
        #
        # # calculate update to the actual gradient concentration:
        # self.pgrad_x = self.pgrad_x - self.alpha*(self.pgrad_x - ceqm_x) * p.dt
        # self.pgrad_y = self.pgrad_y - self.alpha*(self.pgrad_y - ceqm_y) * p.dt
        #
        # sim.rho_pump = 1.0 + self.pgrad_x * (cells.nx_rads/2) + self.pgrad_y * (cells.ny_rads/2)

    def update_channel(self, sim, cells, p):

        cav = 1.0  # concentration at cell centre
        cpi = sim.rho_channel  # concentration at membrane
        z = p.z_channel  # charge of ion
        Do = p.D_membrane  # diffusion constant of ion

        cap = (cav + cpi) / 2  # concentration at midpoint between cell centre and membrane
        cgp = (cpi - cav) / cells.R_rads  # concentration gradients

        # Nernst-Planck equation to calculate flux:
        cfluxco = -Do * cgp + ((Do * p.q * cap * z) / (p.kb * sim.T)) * sim.Ec

        # as no net mass must leave this intracellular movement, make the flux divergence-free:
        cfluxc = stb.single_cell_div_free(cfluxco, cells)

        # calculate the actual concentration at membranes by unpacking to concentration vectors:
        sim.rho_channel = cpi + cfluxc * (cells.mem_sa / cells.mem_vol) * p.dt

        # calculate the equillibrium gradient vector pointing from maximum to minimum density on the membrane:
        # ceqm_x = ((p.z_channel * p.q) / (p.kb * p.T)) * 1.0 * sim.J_cell_x[cells.mem_to_cells] * (1/sim.sigma)
        # ceqm_y = ((p.z_channel * p.q) / (p.kb * p.T)) * 1.0 * sim.J_cell_y[cells.mem_to_cells] * (1/sim.sigma)
        #
        # # calculate update to the actual gradient concentration:
        # self.cgrad_x = self.cgrad_x - self.alpha * (self.cgrad_x - ceqm_x) * p.dt
        # self.cgrad_y = self.cgrad_y - self.alpha * (self.cgrad_y - ceqm_y) * p.dt
        #
        # sim.rho_channel = 1.0 + self.cgrad_x * (cells.nx_rads/2) + self.cgrad_y * (cells.ny_rads/2)

    def remove_data(self, targets_cell):

        pass

        # # remove items from the mem concentration lists:
        # rpo2 = np.delete(self.rpo, targets_cell)
        # # reassign the new data vector to the object:
        # self.rpo = rpo2*1
        #
        # # remove items from the mem concentration lists:
        # rco2 = np.delete(self.rco, targets_cell)
        # # reassign the new data vector to the object:
        # self.rco = rco2*1





#!/usr/bin/env python3
# --------------------( LICENSE                            )--------------------
# Copyright 2014-2025 by Alexis Pietak & Cecil Curry.
# See "LICENSE" for further details.

'''
High-level **simulation pipeline** (i.e., container of related, albeit isolated,
simulation actions to be run iteratively) functionality.
'''

#FIXME: Metadata duplication between pipeline and pipeline runner classes has
#reached Corinthian heights of inanity. To address both this and the BETSEE
#export substatus issue (i.e., that substatus for exports is currently kludged
#at the BETSE layer and hence untranslateable), consider:
#
#* Define a new "betse.util.type.call.callbacks.CallbacksMetadataABC" abstract
#  base class whose body reduces to simply "pass".
#* Refactor the "betse.util.type.call.callbacks.CallbacksBC" class as follows:
#  * Define a new progress_metadated() method accepting a mandatory
#    "metadata : CallbacksMetadataABC" parameter.
#  * Improve the progressed_next() method to accept an optional parameter of
#    the same type.
#* Define a new "SimPipeMetadataABC" abstract base class in this submodule,
#  subclassing "CallbacksMetadataABC" for improved generality.
#  Shift all of the following "SimPipeABC" class properties into
#  "SimPipeMetadataABC" *INSTANCE* (i.e., standard non-class) properties:
#  * "_NOUN_SINGULAR", renamed to "noun_singular_lower".
#  * "_VERB_CONTINUOUS", renamed to "verb_continuous".
#  * Lastly, define a new "noun_singular_upper" property.
#* Define a new SimPipeABC.metadata() abstract class property. All "SimPipeABC"
#  subclasses should redefine this property by:
#  * Defining a new "SimPipeMetadataABC" concrete subclass.
#  * Overriding SimPipeABC.metadata() to return an instance of that subclass.
#* Remove the SimPipeRunnerMetadata.noun* and verb* instance variables.
#* Define a new "SimPipeRunnerMetadata.pipe" instance variable.
#* Refactor the SimPipeABCMeta.__new__() method to set this variable on each
#  pipeline runner.
#* Refactor the SimPipesExport.export() method to pass the optional "metadata"
#  parameter to the call to phase.callbacks.progressed_next(): e.g.,
#      phase.callbacks.progressed_next(metadata=runner_metadata.pipe)
#
#Voila! Frankly trivial, but fairly involved. A real-world dichotomy in action.

# ....................{ IMPORTS                            }....................
from abc import ABCMeta, abstractproperty
from betse.exceptions import BetseSimPipeException
from betse.science.config.export.confexpabc import SimConfExportABC
from betse.science.phase.phasecls import SimPhase
from betse.util.io.log import logs
from betse.util.type.decorator.deccls import abstractmethod #, abstractproperty
from betse.util.type.descriptor.descs import (
    abstractclassproperty_readonly, classproperty_readonly)
from betse.util.type.iterable import itersort
from betse.util.type.obj import objects, objiter, objtest
from betse.util.type.text.string import strs
from betse.util.type.types import (
    type_check,
    ClassType,
    GeneratorType,
    IterableTypes,
    MappingType,
    SequenceTypes,
)

# ....................{ METACLASSES                        }....................
class SimPipeABCMeta(ABCMeta):
    '''
    Metaclass of the abstract :class:`SimPipeABC` base class and all concrete
    subclasses thereof.

    This metaclass dynamically annotates each **simulation pipeline runner**
    (i.e., method decorated by the :func:`piperunner` decorator) on the initial
    definition of each concrete subclass of the :class:`SimPipeABC` superclass
    with metadata pertaining to that subclass, augmenting the core
    runner-specific metadata already annotated on that runner by that
    decorator. (The metadata annotated by this metaclass requires access to the
    class declaring that runner, which the :func:`piperunner` decorator has no
    means of accessing under conventional semantics of method declaration.)

    See Also
    ----------
    https://stackoverflow.com/a/47371003/2809027
        StackOverflow answer mildly inspiring this class.
    '''

    # ..................{ CONSTRUCTORS                       }..................
    def __new__(
        metacls: ClassType,
        class_name: str,
        class_base_classes: IterableTypes,
        class_attrs: MappingType,
        **kwargs
    ) -> ClassType:
        '''
        Annotate each simulation pipeline runner defined by the passed
        simulation pipeline subclass with metadata pertaining to that subclass.
        '''

        # Unsanitized "SimPipeABC" subclass.
        subcls = super().__new__(
            metacls, class_name, class_base_classes, class_attrs)

        # For the method name and method of each runner in this pipeline...
        for runner_method_name, runner_method in subcls.iter_runners_method():
            # Metadata associated with this runner.
            runner_metadata = runner_method.metadata

            # Set this runner's type to the name of this method excluding this
            # pipeline's runner prefix.
            runner_metadata.kind = strs.remove_prefix(
                text=runner_method_name,
                prefix=subcls._RUNNER_METHOD_NAME_PREFIX,
            )

            # Set this runner's human-readable type to that of this pipeline.
            runner_metadata.noun_singular_lowercase = subcls._NOUN_SINGULAR
            runner_metadata.noun_singular_uppercase = (
                subcls._NOUN_SINGULAR.upper())
            runner_metadata.verb_continuous = subcls._VERB_CONTINUOUS

        # Return this sanitized "SimPipeABC" subclass.
        return subcls

# ....................{ SUPERCLASSES                       }....................
#FIXME: Revise docstring to account for the recent large-scale class redesign.
class SimPipeABC(object, metaclass=SimPipeABCMeta):
    '''
    Abstract base class of all subclasses running a **simulation pipeline**
    (i.e., sequence of similar simulation activities to be iteratively run).

    This class implements the infrastructure for iteratively running all
    currently enabled simulation activities (referred to as "runners")
    comprising the pipeline defined by this subclass, either in parallel *or*
    in series.

    Design
    ----------
    Each subclass is expected to define:

    * One or more public methods with names prefixed by the subclass-specific
      :attr:`_RUNNER_METHOD_NAME_PREFIX`, defaulting to ``run_`` (e.g.,
      ``run_voltage_intra``). For each such method, the name of that method
      excluding that prefix is the name of that method's runner (e.g.,
      ``voltage_intra`` for the method name ``run_voltage_intra``). Each such
      method should:

      * Accept exactly two parameters:

        * ``self``.
        * ``conf``, an object supplying this runner with its configuration.
          This is typically but *not* necessarily an instance of a subclass of
          the :class:`betse.lib.yaml.abc.yamlabc.YamlABC` superclass, enabling
          transparent configuration of runners from YAML-based files.

      * Return nothing (i.e.,``None``).

    * The abstract :meth:`iter_runners_conf` method returning a sequence of the
      names of all runners currently enabled by this pipeline (e.g.,
      ``['voltage_intra', 'ion_hydrogen', 'electric_total']``).

    The :meth:`run` method defined by this base class then dynamically
    implements this pipeline by iterating over the :meth:`iter_runners_conf`
    property and, for each enabled runner, calling that runner's method.

    Attributes (Private: Labels)
    ----------
    _noun_singular_lowercase : str
        Human-readable lowercase singular noun synopsizing the type of runners
        implemented by this subclass (e.g., ``animation``, ``plot``).
    _noun_singular_uppercase : str
        Human-readable singular noun whose first character is uppercase and all
        remaining characters lowercase (e.g., ``Animation``, ``Plot``).
    _noun_plural_lowercase : str
        Human-readable lowercase plural noun synopsizing the type of runners
        implemented by this subclass (e.g., ``animations``, ``plots``).
    '''

    # ..................{ SUBCLASS ~ properties              }..................
    # Subclasses are required to implement the following abstract properties.

    @abstractclassproperty_readonly
    def _NOUN_SINGULAR(cls) -> str:
        '''
        Human-readable singular noun synopsizing the type of runners
        implemented by this subclass (e.g., ``animation``, ``plot``), ideally
        but *not* necessarily lowercase.
        '''

        # Return a nonsensical placeholder constant to satisfy @beartype. *sigh*
        return '"SimPipeABC._NOUN_SINGULAR" undefined'
        # raise NotImplementedError(
        #     f'Simulation pipe subclass {repr(cls)} '
        #     f'abstract property "_NOUN_SINGULAR" unimplemented.'
        # )

    # ..................{ CLASS ~ properties : concrete      }..................
    @classproperty_readonly
    def _NOUN_PLURAL(cls) -> str:
        '''
        Human-readable plural noun synopsizing the type of runners implemented
        by this subclass (e.g., ``animations``, ``plots``), ideally but *not*
        necessarily lowercase.

        Defaults to suffixing :meth:`_NOUN_SINGULAR` by ``s``.
        '''

        return cls._NOUN_SINGULAR + 's'


    @classproperty_readonly
    def _VERB_CONTINUOUS(cls) -> str:
        '''
        Human-readable verb in the continuous tense synopsizing the type of
        action performed by runners implemented by this subclass (e.g.,
        ``Saving``), ideally but *not* necessarily capitalized.

        Defaults to ``Running``.
        '''

        return 'Running'


    @classproperty_readonly
    def _RUNNER_METHOD_NAME_PREFIX(cls) -> str:
        '''
        Substring prefixing the name of each runner defined by this pipeline.

        Defaults to ``run_``.
        '''

        return  'run_'

    # ..................{ CLASS ~ iterators                  }..................
    @classmethod
    def iter_runners_method(cls) -> GeneratorType:
        '''
        Generator yielding the method name and method of each **simulation
        pipeline runner** (i.e., method bound to this pipeline decorated by the
        :func:`piperunner` decorator).

        This generator excludes all methods defined by this pipeline subclass
        *not* decorated by that decorator.

        Yields
        ----------
        (runner_method_name : str, runner_method : MethodType)
            2-tuple where:

            * ``runner_method_name`` is the name of the method underlying this
              runner, identical to the ``runner_method.__name__`` attribute.
            * ``runner_method`` is this method.
        '''

        # Defer to this generator.
        yield from objiter.iter_methods_prefixed(
            obj=cls, prefix=cls._RUNNER_METHOD_NAME_PREFIX)

    # ..................{ CLASS ~ iterators : metadata       }..................
    @classmethod
    def iter_runners_metadata(cls) -> GeneratorType:
        '''
        Generator yielding the metadata annotating each **simulation pipeline
        runner** (i.e., method bound to this pipeline decorated by the
        :func:`piperunner` decorator).

        This generator excludes all methods defined by this pipeline subclass
        *not* decorated by that decorator.

        Yields
        ----------
        runner_metadata : SimPipeRunnerMetadata
            :class:`SimPipeRunnerMetadata` instance collecting all metadata for
            this runner.
        '''

        # Return a generator comprehension iteratively yielding...
        return (
            # Metadata annotating this runner.
            runner_method.metadata
            # For each runner method defined by this pipeline subclass...
            for _, runner_method in cls.iter_runners_method()
        )


    @classmethod
    @type_check
    def iter_runners_metadata_kind(cls) -> SequenceTypes:
        '''
        Sequence of each **machine-readable type** (i.e., string value of the
        :attr:`SimPipeRunnerMetadata.kind` instance variable) of all runners
        supported by this pipeline (in sorted lexicographic order).
        '''

        # Return a lexicographically sorted sequence of...
        return itersort.sort_ascending(
            # The machine-readable type of this runner.
            runner_metadata.kind
            # For each runner defined by this pipeline subclass...
            for runner_metadata in cls.iter_runners_metadata()
        )

    # ..................{ INITIALIZERS                       }..................
    @type_check
    def __init__(self) -> None:
        '''
        Initialize this pipeline.
        '''

        # Classify all passed parameters.
        self._noun_singular_lowercase = self._NOUN_SINGULAR
        self._noun_singular_uppercase = strs.uppercase_char_first(
            self._NOUN_SINGULAR)
        self._noun_plural_lowercase = self._NOUN_PLURAL


    @type_check
    def init(self, phase: SimPhase) -> None:
        '''
        Initialize this pipeline for the passed simulation phase.

        Defaults to a noop. Pipeline subclasses may override this method to
        guarantee state assumed by subsequently run pipeline runners (e.g., to
        create external directories required by these runners for this phase).

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        pass

    # ..................{ SUBCLASS ~ methods                 }..................
    # Subclasses are required to implement the following abstract methods.

    @abstractmethod
    def iter_runners_conf(self, phase: SimPhase) -> SequenceTypes:
        '''
        Sequence of all **runner configurations** (i.e.,
        :class:`SimConfExportABC` instances encapsulating all input parameters
        passed to the corresponding pipeline runner) for the passed simulation
        phase.

        Pipeline subclasses typically implement this property to return an
        instance of the :class:``YamlList`` class listing all runners listed
        by the simulation configuration file configuring this phase.

        Caveats
        ----------
        The existence of a runner configuration does *not* imply the
        corresponding pipeline runner to be unconditionally enabled. Instead,
        the :attr:`SimConfExportABC.is_enabled` data descriptor defined by all
        runner configurations returned by this method specifies whether or not
        that runner is to be enabled or disabled.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        pass

    # ..................{ PROPERTIES                         }..................
    @property
    def name(self) -> str:
        '''
        Human-readable name of this pipeline, intended principally for display
        (e.g., logging) purposes.

        Defaults to suffixing :meth:`_NOUN_SINGULAR` by `` pipeline``.
        '''

        return '{} pipeline'.format(self._NOUN_SINGULAR)

    # ..................{ PROPERTIES ~ private               }..................
    @type_check
    def _is_enabled(self, phase: SimPhase) -> bool:
        '''
        ``True`` only if this pipeline is enabled for the passed simulation
        phase.

        Specifically, if this boolean is:

        * ``False``, the :meth:`run` method reduces to a noop.
        * ``True``, that method behaves as expected (i.e., calls all currently
          enabled runner methods).

        Defaults to ``True``. Pipeline subclasses typically override this
        property to return a boolean derived from the simulation configuration
        file associated with the passed phase.

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.
        '''

        return True

    # ..................{ ITERATORS                          }..................
    @type_check
    def iter_runners_enabled(self, phase: SimPhase) -> GeneratorType:
        '''
        Generator yielding the method and simulation configuration of each
        **enabled simulation pipeline runner** (i.e., method bound to this
        pipeline decorated by the :func:`piperunner` decorator *and* enabled by
        this simulation configuration) for the passed simulation phase.

        If this simulation phase disables this pipeline, this generator reduces
        to the empty generator (i.e., yields no values); else, this generator
        excludes:

        * All methods defined by this pipeline subclass *not* decorated by the
          :func:`piperunner` decorator.
        * All methods defined by this pipeline subclass decorated by that
          decorator currently disabled by their simulation configuration (e.g.,
          such that the YAML-backed value of the ``enabled`` key in this
          configuration evaluates to ``False``).

        Parameters
        ----------
        phase : SimPhase
            Current simulation phase.

        Yields
        ----------
        (runner_method : MethodType, runner_conf : SimConfExportABC)
            2-tuple where:

            * ``runner_method`` is the method implementing this runner, whose
              method name is guaranteed to be prefixed by the substring
              :attr:`_RUNNER_METHOD_NAME_PREFIX`. Note that this method object
              defines the following custom instance variables:

              * ``metadata``, whose value is an instance of the
                :class:`SimPipeRunnerMetadata` class.

            * ``runner_conf`` is the YAML-backed list item configuring this
              runner, derived from the YAML-formatted simulation configuration
              file associated with the passed simulation phase.

        Raises
        ----------
        BetseSimPipeException
            If any simulation runner configured for this pipeline by the passed
            simulation phase is **unrecognized** (i.e., if this pipeline
            defines no corresponding method).
        '''

        # If this pipeline is disabled, log this fact and return immediately.
        if not self._is_enabled(phase):
            logs.log_debug(
                'Ignoring disabled %s...', self._noun_plural_lowercase)
            return
        # Else, this pipeline is enabled.

        # For each runner configuration specified for this pipeline...
        for runner_conf in self.iter_runners_conf(phase):
            # If this configuration is *NOT* YAML-backed, raise an exception.
            objtest.die_unless_instance(obj=runner_conf, cls=SimConfExportABC)

            # If this runner is disabled, log this fact and ignore this runner.
            if not runner_conf.is_enabled:
                logs.log_debug(
                    'Ignoring disabled %s "%s"...',
                    self._noun_singular_lowercase,
                    runner_conf.kind)
                continue
            # Else, this runner is enabled.

            # Name of the pipeline method implementing this runner.
            runner_method_name = (
                self._RUNNER_METHOD_NAME_PREFIX + runner_conf.kind)

            # Method running this runner if recognized *OR* "None" otherwise.
            runner_method = objects.get_callable_or_none(
                obj=self, callable_name=runner_method_name)

            # If this runner is unrecognized, raise an exception.
            if runner_method is None:
                raise BetseSimPipeException(
                    '{} "{}" unrecognized '
                    '(i.e., method {}.{}() not found).'.format(
                        self._noun_singular_uppercase,
                        runner_conf.kind,
                        objects.get_class_name_unqualified(self),
                        runner_method_name))
            # Else, this runner is recognized.

            # Yield a 2-tuple of this runner's method and configuration.
            yield runner_method, runner_conf
